/****************************************************************************
 * Visual Boy Advance GX
 *
 * Tantric September 2008
 *
 * fileop.h
 *
 * File operations
 ****************************************************************************/

#ifndef _FILEOP_H_
#define _FILEOP_H_

#include <gccore.h>
#include <stdio.h>
#include <string.h>
#include <ogcsys.h>
#include <fat.h>
#include <unistd.h>

#define SAVEBUFFERSIZE (1024 * 1024 * 2)

void InitDeviceThread();
void ResumeDeviceThread();
void HaltDeviceThread();
void HaltParseThread();
void MountAllFAT();
void UnmountAllFAT();
bool FindDevice(char * filepath, int * device);
char * StripDevice(char * path);
bool ChangeInterface(int device, bool silent);
bool ChangeInterface(char * filepath, bool silent);
void CreateAppPath(char * origpath);
int ParseDirectory(bool waitParse = false, bool filter = true);
void AllocSaveBuffer();
void FreeSaveBuffer();
size_t LoadFile(char * rbuffer, char *filepath, size_t length, bool silent);
size_t LoadFile(char * filepath, bool silent);
size_t LoadSzFile(char * filepath, unsigned char * rbuffer);
size_t SaveFile(char * buffer, char *filepath, size_t datasize, bool silent);
size_t SaveFile(char * filepath, size_t datasize, bool silent);

extern unsigned char *savebuffer;
extern FILE * file;
extern bool unmountRequired[];
extern bool isMounted[];
extern int selectLoadedFile;


//this is a good enough spot for these
s32 WDVD_Init(void);
s32 WDVD_OpenDataPartition();
s32 WDVD_Close(void);
s32 WDVD_Read(void *buf, u32 len, u64 offset);
s32 WDVD_Read_Safe(void *buf, u32 len, u64 offset);
s32 WDVD_ClosePartition(void);
s32 WDVD_GetMainDol( void *outbuf, u32* size );

bool FSTIsMounted();
bool FST_Mount();
bool FST_Unmount();

//simple debugging.  this seems like a good enough place to put this
void gprintf(const char *str, ...);
void hexdump(void *d, int len);

#endif
