/****************************************************************************
 * Visual Boy Advance GX
 *
 * Tantric September 2008
 *
 * vbagx.cpp
 *
 * This file controls overall program flow. Most things start and end here!
 ***************************************************************************/

#include <gccore.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ogcsys.h>
#include <unistd.h>
#include <wiiuse/wpad.h>
#include <sys/iosupport.h>

//#ifdef HW_RVL
//#include <di/di.h>
//#endif

#include "vbagx.h"
#include "vbasupport.h"
#include "preferences.h"
#include "audio.h"
#include "networkop.h"
#include "filebrowser.h"
#include "fileop.h"
#include "menu.h"
#include "input.h"
#include "video.h"
#include "gamesettings.h"
#include "mem2.h"
#include "utils/FreeTypeGX.h"

#include "vba/gba/Globals.h"
#include "vba/gba/Sound.h"

extern "C" {
extern void __exception_setreload(int t);
extern u32 __di_check_ahbprot(void);
}

//variables for "rebooting" the iso
static u8 *homebrewbuffer = (u8 *)0x92000000;
u32 homebrewsize = 0;
int BootHomebrew();
void AddBootArgument(const char * argv);
extern const u8		app_booter_dol[];
extern const u32	app_booter_dol_size;
extern "C" {
extern void __exception_closeall();
typedef void (*entrypoint) (void);
}

extern int emulating;
void StopColorizing();
void gbSetPalette(u32 RRGGBB[]);
int ScreenshotRequested = 0;
int ConfigRequested = 0;
int ShutdownRequested = 0;
int ResetRequested = 0;
int ExitRequested = 0;
char appPath[1024] = { 0 };
char loadedFile[1024] = { 0 };

//#define DBG

/****************************************************************************
 * Shutdown / Reboot / Exit
 ***************************************************************************/

static void ExitCleanup()
{
	ShutdownAudio();
	StopGX();
	HaltDeviceThread();
	UnmountAllFAT();
	FST_Unmount();

//#ifdef HW_RVL
//	DI_Close();
//#endif
}

#ifdef HW_DOL
	#define PSOSDLOADID 0x7c6000a6
	int *psoid = (int *) 0x80001800;
	void (*PSOReload) () = (void (*)()) 0x80001800;
#endif

void ExitApp()
{
#ifdef HW_RVL
	ShutoffRumble();
	WPAD_Shutdown();
#endif

	SavePrefs(SILENT);

	if (ROMLoaded && !ConfigRequested && GCSettings.AutoSave == 1)
		SaveBatteryOrStateAuto(FILE_SRAM, SILENT);

	ExitCleanup();

	if(ShutdownRequested)
		SYS_ResetSystem(SYS_POWEROFF, 0, 0);

#ifdef DBG
    exit( 0 );
    while( 1 );
#endif

	s32 hb = WDVD_GetMainDol( (void*)homebrewbuffer, &homebrewsize );
	WDVD_ClosePartition();
	WDVD_Close();
	if( !hb )
	{
	    AddBootArgument( "dvd:/main.dol" );
	    AddBootArgument( "dvd:/gba.dol" );
	    BootHomebrew();
	}
	SYS_ResetSystem(SYS_RETURNTOMENU, 0, 0);

	/*#ifdef HW_RVL
	if(GCSettings.ExitAction == 0) // Auto
	{
		char * sig = (char *)0x80001804;
		if(
			sig[0] == 'S' &&
			sig[1] == 'T' &&
			sig[2] == 'U' &&
			sig[3] == 'B' &&
			sig[4] == 'H' &&
			sig[5] == 'A' &&
			sig[6] == 'X' &&
			sig[7] == 'X')
			GCSettings.ExitAction = 3; // Exit to HBC
		else
			GCSettings.ExitAction = 1; // HBC not found
	}
	#endif

	if(GCSettings.ExitAction == 1) // Exit to Menu
	{
		#ifdef HW_RVL
			SYS_ResetSystem(SYS_RETURNTOMENU, 0, 0);
		#else
			#define SOFTRESET_ADR ((volatile u32*)0xCC003024)
			*SOFTRESET_ADR = 0x00000000;
		#endif
	}
	else if(GCSettings.ExitAction == 2) // Shutdown Wii
	{
		SYS_ResetSystem(SYS_POWEROFF, 0, 0);
	}
	else // Exit to Loader
	{
		#ifdef HW_RVL
			exit(0);
		#else
			if (psoid[0] == PSOSDLOADID)
				PSOReload();
		#endif
	}*/
}

#ifdef HW_RVL
void ShutdownCB()
{
	ShutdownRequested = 1;
}
void ResetCB()
{
	ResetRequested = 1;
}
#endif

#ifdef HW_DOL
/****************************************************************************
 * ipl_set_config
 * lowlevel Qoob Modchip disable
 ***************************************************************************/

static void ipl_set_config(unsigned char c)
{
	volatile unsigned long* exi = (volatile unsigned long*)0xCC006800;
	unsigned long val,addr;
	addr=0xc0000000;
	val = c << 24;
	exi[0] = ((((exi[0]) & 0x405) | 256) | 48);	//select IPL
	//write addr of IPL
	exi[0 * 5 + 4] = addr;
	exi[0 * 5 + 3] = ((4 - 1) << 4) | (1 << 2) | 1;
	while (exi[0 * 5 + 3] & 1);
	//write the ipl we want to send
	exi[0 * 5 + 4] = val;
	exi[0 * 5 + 3] = ((4 - 1) << 4) | (1 << 2) | 1;
	while (exi[0 * 5 + 3] & 1);

	exi[0] &= 0x405;	//deselect IPL
}
#endif

/****************************************************************************
 * USB Gecko Debugging
 ***************************************************************************/

static bool gecko = false;
static mutex_t gecko_mutex = 0;

static ssize_t __out_write(struct _reent *r, int fd, const char *ptr, size_t len)
{
	u32 level;

	if (!ptr || len <= 0 || !gecko)
		return -1;

	LWP_MutexLock(gecko_mutex);
	level = IRQ_Disable();
	usb_sendbuffer(1, ptr, len);
	IRQ_Restore(level);
	LWP_MutexUnlock(gecko_mutex);
	return len;
}

const devoptab_t gecko_out = {
	"stdout",	// device name
	0,			// size of file structure
	NULL,		// device open
	NULL,		// device close
	__out_write,// device write
	NULL,		// device read
	NULL,		// device seek
	NULL,		// device fstat
	NULL,		// device stat
	NULL,		// device link
	NULL,		// device unlink
	NULL,		// device chdir
	NULL,		// device rename
	NULL,		// device mkdir
	0,			// dirStateSize
	NULL,		// device diropen_r
	NULL,		// device dirreset_r
	NULL,		// device dirnext_r
	NULL,		// device dirclose_r
	NULL		// device statvfs_r
};

void USBGeckoOutput()
{
	LWP_MutexInit(&gecko_mutex, false);
	gecko = usb_isgeckoalive(1);
	
	devoptab_list[STD_OUT] = &gecko_out;
	devoptab_list[STD_ERR] = &gecko_out;
}

/****************************************************************************
* main
*
* Program entry
****************************************************************************/
int main(int argc, char *argv[])
{
#ifdef DBG
	IOS_ReloadIOS( 35 ); //only for testing
#endif
	USBGeckoOutput(); // uncomment to enable USB gecko output
	__exception_setreload(0);

	#ifdef HW_DOL
	ipl_set_config(6); // disable Qoob modchip
	#endif

#ifdef HW_RVL
	// only reload IOS if AHBPROT is not enabled
	/*u32 version = IOS_GetVersion();
	s32 preferred = IOS_GetPreferredVersion();

	if(version != 58 && preferred > 0 && version != (u32)preferred && __di_check_ahbprot() != 1)
		IOS_ReloadIOS(preferred);
*/
	StartNetworkThread();
	//DI_Init();
	WDVD_Init();
	WDVD_OpenDataPartition();
#endif

	InitDeviceThread();
	InitializeVideo();
	ResetVideo_Menu (); // change to menu video mode
	SetupPads();
	
	// Initialize DVD subsystem (GameCube only)
	#ifdef HW_DOL
	DVD_Init ();
	#endif

	#ifdef HW_RVL
	// Wii Power/Reset buttons
	WPAD_SetPowerButtonCallback((WPADShutdownCallback)ShutdownCB);
	SYS_SetPowerCallback(ShutdownCB);
	SYS_SetResetCallback(ResetCB);
	#endif

	MountAllFAT(); // Initialize libFAT for SD and USB

	InitialiseSound();
	InitialisePalette();
	DefaultSettings (); // Set defaults
	InitFreeType((u8*)font_ttf, font_ttf_size); // Initialize font system
#ifdef HW_RVL
	if(argc > 0 && argv[0] != NULL)
		CreateAppPath(argv[0]); // store path app was loaded from

	InitMem2Manager();
	savebuffer = (unsigned char *)mem2_malloc(SAVEBUFFERSIZE);
	browserList = (BROWSERENTRY *)mem2_malloc(sizeof(BROWSERENTRY)*MAX_BROWSER_SIZE);
	rom = (u8 *)mem2_malloc(1024*1024*32); // allocate 32 MB to GBA ROM
#else
	savebuffer = (unsigned char *)malloc(SAVEBUFFERSIZE);
	browserList = (BROWSERENTRY *)malloc(sizeof(BROWSERENTRY)*MAX_BROWSER_SIZE);
#endif
	gameScreenPng = (u8 *)malloc(512*1024);
	InitGUIThreads();

	while(1) // main loop
	{
		// go back to checking if devices were inserted/removed
		// since we're entering the menu
		ResumeDeviceThread();

		SwitchAudioMode(1);

		if(!ROMLoaded)
			MainMenu(MENU_GAMESELECTION);
		else
			MainMenu(MENU_GAME);

		ConfigRequested = 0;
		ScreenshotRequested = 0;
		SwitchAudioMode(0);

		// stop checking if devices were removed/inserted
		// since we're starting emulation again
		HaltDeviceThread();

		ResetVideo_Emu();

		// GB colorizing - set palette
		if(IsGameboyGame())
		{
			if(GCSettings.colorize && strcmp(RomTitle, "MEGAMAN") != 0)
				gbSetPalette(CurrentPalette.palette);
			else
				StopColorizing();
		}

		while (emulating) // emulation loop
		{
			emulator.emuMain(emulator.emuCount);

			if(ResetRequested)
			{
				emulator.emuReset(); // reset game
				ResetRequested = 0;
			}
			if(ConfigRequested)
			{
				ResetVideo_Menu();
				break; // leave emulation loop
			}
			#ifdef HW_RVL
			if(ShutdownRequested)
				ExitApp();
			#endif
		} // emulation loop
	} // main loop
	return 0;
}

#include <ogc/machine/processor.h>
#include <vector>
#include <string>

typedef struct _dolheader {
    u32 text_pos[7];
    u32 data_pos[11];
    u32 text_start[7];
    u32 data_start[11];
    u32 text_size[7];
    u32 data_size[11];
    u32 bss_start;
    u32 bss_size;
    u32 entry_point;
} dolheader;

u32 load_dol(const void *dolstart, struct __argv *argv) {
    u32 i;
    dolheader *dolfile;

    if (dolstart) {
	dolfile = (dolheader *) dolstart;
	for (i = 0; i < 7; i++) {
	    if ((!dolfile->text_size[i]) || (dolfile->text_start[i] < 0x100)) continue;
	    ICInvalidateRange ((void *) dolfile->text_start[i],dolfile->text_size[i]);
	    memcpy((void *) dolfile->text_start[i],dolstart+dolfile->text_pos[i],dolfile->text_size[i]);
	}

	for (i = 0; i < 11; i++) {
	    if ((!dolfile->data_size[i]) || (dolfile->data_start[i] < 0x100)) continue;
	    memcpy((void *) dolfile->data_start[i],dolstart+dolfile->data_pos[i],dolfile->data_size[i]);
	    DCFlushRangeNoSync ((void *) dolfile->data_start[i],dolfile->data_size[i]);
	}

	memset ((void *) dolfile->bss_start, 0, dolfile->bss_size);
	DCFlushRange((void *) dolfile->bss_start, dolfile->bss_size);

	if (argv && argv->argvMagic == ARGV_MAGIC) {
	    void *new_argv = (void *)(dolfile->entry_point + 8);
	    memcpy(new_argv, argv, sizeof(*argv));
	    DCFlushRange(new_argv, sizeof(*argv));
	}

	return dolfile->entry_point;
    }
    return 0;
}

static std::vector<std::string> Arguments;

void AddBootArgument(const char * argv)
{
    std::string arg(argv);
    Arguments.push_back(arg);
}

static int SetupARGV(struct __argv * args)
{
    if(!args)
	return -1;

    bzero(args, sizeof(struct __argv));
    args->argvMagic = ARGV_MAGIC;

    u32 stringlength = 1;

    for(u32 i = 0; i < Arguments.size(); i++)
    {
	stringlength += Arguments[i].size()+1;
    }

    args->length = stringlength;
    args->commandLine = (char*) malloc(args->length);
    if (!args->commandLine)
	return -1;

    u32 argc = 0;
    u32 position = 0;

    for(u32 i = 0; i < Arguments.size(); i++)
    {
	strcpy(&args->commandLine[position], Arguments[i].c_str());
	position += Arguments[i].size()+1;
	argc++;
    }

    args->argc = argc;

    args->commandLine[args->length - 1] = '\0';
    args->argv = &args->commandLine;
    args->endARGV = args->argv + 1;

    Arguments.clear();

    return 0;
}

int BootHomebrew()
{
    if( homebrewsize == 0 )
	return -1;

    struct __argv args;
    SetupARGV(&args);

    u32 cpu_isr;

    entrypoint entry = (entrypoint)load_dol( app_booter_dol, &args );

    if ( !entry )
	return -1;

    SYS_ResetSystem( SYS_SHUTDOWN, 0, 0 );
    _CPU_ISR_Disable( cpu_isr );
    __exception_closeall();
    entry();
    _CPU_ISR_Restore ( cpu_isr );

    return 0;
}
