/****************************************************************************
 * libwiigui Template
 * Tantric 2009
 *
 * menu.cpp
 * Menu flow routines - handles all menu logic
 ***************************************************************************/

#include <gccore.h>
#include <ogcsys.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <wiiuse/wpad.h>
#include <sdcard/wiisd_io.h>

#include "libwiigui/gui.h"
#include "menu.h"
#include "demo.h"
#include "input.h"
#include "filelist.h"
#include "filebrowser.h"
#include "di_shit.h"
#include "BootHomebrew/BootHomebrew.h"

#define THREAD_SLEEP 100

#define N64_PATH "dvd:/n64.dol"
#define NES_PATH "dvd:/nes.dol"
#define SEGA_PATH "dvd:/sega.dol"
#define SNES_PATH "dvd:/snes.dol"
#define GBA_PATH "dvd:/gba.dol"

#define N64_SD_PATH "sd:/wii64/boot.dol"
#define NES_SD_PATH "sd:/fcuegx/boot.dol"
#define SEGA_SD_PATH "sd:/genplus/boot.dol"
#define SNES_SD_PATH "sd:/snes9xgx/boot.dol"
#define GBA_SD_PATH "sd:/vbagx/boot.dol"

#define N64_USB_PATH "usb:/wii64/boot.dol"
#define NES_USB_PATH "usb:/fcuegx/boot.dol"
#define SEGA_USB_PATH "usb:/genplus/boot.dol"
#define SNES_USB_PATH "usb:/snes9xgx/boot.dol"
#define GBA_USB_PATH "usb:/vbagx/boot.dol"

static GuiImageData * pointer[4];
static GuiImage * bgImg = NULL;
static GuiSound * bgMusic = NULL;
static GuiWindow * mainWindow = NULL;
static lwp_t guithread = LWP_THREAD_NULL;
static bool guiHalt = true;

char lastEmu[ 64 ] = { 0 };
static bool firstRun = false;

int region = -1;
int widescreen = 0;

/****************************************************************************
 * ResumeGui
 *
 * Signals the GUI thread to start, and resumes the thread. This is called
 * after finishing the removal/insertion of new elements, and after initial
 * GUI setup.
 ***************************************************************************/
static void
ResumeGui()
{
	guiHalt = false;
	LWP_ResumeThread (guithread);
}

/****************************************************************************
 * HaltGui
 *
 * Signals the GUI thread to stop, and waits for GUI thread to stop
 * This is necessary whenever removing/inserting new elements into the GUI.
 * This eliminates the possibility that the GUI is in the middle of accessing
 * an element that is being changed.
 ***************************************************************************/
static void
HaltGui()
{
	guiHalt = true;

	// wait for thread to finish
	while(!LWP_ThreadIsSuspended(guithread))
		usleep(THREAD_SLEEP);
}

/****************************************************************************
 * WindowPrompt
 *
 * Displays a prompt window to user, with information, an error message, or
 * presenting a user with a choice
 ***************************************************************************/
int
WindowPrompt(const char *title, const char *msg, const char *btn1Label, const char *btn2Label)
{
	int choice = -1;

	GuiWindow promptWindow(448,288);
	promptWindow.SetAlignment(ALIGN_CENTRE, ALIGN_MIDDLE);
	promptWindow.SetPosition(0, -10);
	GuiSound btnSoundOver(button_over_pcm, button_over_pcm_size, SOUND_PCM);
	GuiImageData btnOutline(button_png);
	GuiImageData btnOutlineOver(button_over_png);
	GuiTrigger trigA;
	trigA.SetSimpleTrigger(-1, WPAD_BUTTON_A | WPAD_CLASSIC_BUTTON_A, PAD_BUTTON_A);

	GuiImageData dialogBox(dialogue_box_png);
	GuiImage dialogBoxImg(&dialogBox);

	GuiText titleTxt(title, 24, (GXColor){0, 0, 0, 255});
	titleTxt.SetAlignment(ALIGN_CENTRE, ALIGN_TOP);
	titleTxt.SetPosition(0,40);
	GuiText msgTxt(msg, 22, (GXColor){0, 0, 0, 255});
	msgTxt.SetAlignment(ALIGN_CENTRE, ALIGN_MIDDLE);
	msgTxt.SetPosition(0,-20);
	msgTxt.SetWrap(true, 400);

	GuiText btn1Txt(btn1Label, 22, (GXColor){0, 0, 0, 255});
	GuiImage btn1Img(&btnOutline);
	GuiImage btn1ImgOver(&btnOutlineOver);
	GuiButton btn1(btnOutline.GetWidth(), btnOutline.GetHeight());

	if(btn2Label)
	{
		btn1.SetAlignment(ALIGN_LEFT, ALIGN_BOTTOM);
		btn1.SetPosition(20, -25);
	}
	else
	{
		btn1.SetAlignment(ALIGN_CENTRE, ALIGN_BOTTOM);
		btn1.SetPosition(0, -25);
	}

	btn1.SetLabel(&btn1Txt);
	btn1.SetImage(&btn1Img);
	btn1.SetImageOver(&btn1ImgOver);
	btn1.SetSoundOver(&btnSoundOver);
	btn1.SetTrigger(&trigA);
	btn1.SetState(STATE_SELECTED);
	btn1.SetEffectGrow();

	GuiText btn2Txt(btn2Label, 22, (GXColor){0, 0, 0, 255});
	GuiImage btn2Img(&btnOutline);
	GuiImage btn2ImgOver(&btnOutlineOver);
	GuiButton btn2(btnOutline.GetWidth(), btnOutline.GetHeight());
	btn2.SetAlignment(ALIGN_RIGHT, ALIGN_BOTTOM);
	btn2.SetPosition(-20, -25);
	btn2.SetLabel(&btn2Txt);
	btn2.SetImage(&btn2Img);
	btn2.SetImageOver(&btn2ImgOver);
	btn2.SetSoundOver(&btnSoundOver);
	btn2.SetTrigger(&trigA);
	btn2.SetEffectGrow();

	promptWindow.Append(&dialogBoxImg);
	promptWindow.Append(&titleTxt);
	promptWindow.Append(&msgTxt);
	promptWindow.Append(&btn1);

	if(btn2Label)
		promptWindow.Append(&btn2);

	promptWindow.SetEffect(EFFECT_SLIDE_TOP | EFFECT_SLIDE_IN, 50);
	HaltGui();
	mainWindow->SetState(STATE_DISABLED);
	mainWindow->Append(&promptWindow);
	mainWindow->ChangeFocus(&promptWindow);
	ResumeGui();

	while(choice == -1)
	{
		usleep(THREAD_SLEEP);

		if(btn1.GetState() == STATE_CLICKED)
			choice = 1;
		else if(btn2.GetState() == STATE_CLICKED)
			choice = 0;
	}

	promptWindow.SetEffect(EFFECT_SLIDE_TOP | EFFECT_SLIDE_OUT, 50);
	while(promptWindow.GetEffect() > 0) usleep(THREAD_SLEEP);
	HaltGui();
	mainWindow->Remove(&promptWindow);
	mainWindow->SetState(STATE_DEFAULT);
	ResumeGui();
	return choice;
}

/****************************************************************************
 * UpdateGUI
 *
 * Primary thread to allow GUI to respond to state changes, and draws GUI
 ***************************************************************************/

static void *
UpdateGUI (void *arg)
{
	int i;

	while(1)
	{
		if(guiHalt)
		{
			LWP_SuspendThread(guithread);
		}
		else
		{
			UpdatePads();
			mainWindow->Draw();

			#ifdef HW_RVL
			for(i=3; i >= 0; i--) // so that player 1's cursor appears on top!
			{
				if(userInput[i].wpad->ir.valid)
					Menu_DrawImg(userInput[i].wpad->ir.x-48, userInput[i].wpad->ir.y-48,
						     96, 96, pointer[i]->GetImage(), userInput[i].wpad->ir.angle, widescreen ? .80 : 1, 1, 255);
				DoRumble(i);
			}
			#endif

			Menu_Render();

			for(i=0; i < 4; i++)
				mainWindow->Update(&userInput[i]);

			if(ExitRequested)
			{
				for(i = 0; i <= 255; i += 15)
				{
					mainWindow->Draw();
					Menu_DrawRectangle(0,0,screenwidth,screenheight,(GXColor){0, 0, 0, i},1);
					Menu_Render();
				}
				ExitApp();
			}
		}
	}
	return NULL;
}

/****************************************************************************
 * InitGUIThread
 *
 * Startup GUI threads
 ***************************************************************************/
void
InitGUIThreads()
{
	LWP_CreateThread (&guithread, UpdateGUI, NULL, NULL, 0, 70);
}

bool CheckFile( const char filepath[] )
{
    FILE *f = fopen( filepath, "rb" );
    if( !f )
    {
	return false;
    }

    fclose( f );

    return true;
}

bool HaveSD()
{
    if( !__io_wiisd.startup() )
	    return false;

    bool ret = __io_wiisd.isInserted();

    __io_wiisd.shutdown();

    return ret;
}
//#define DBG_LOCATIONS
static int MenuEmuSelector()
{
	int menu = MENU_NONE;

	GuiSound btnSoundOver( button_over_pcm, button_over_pcm_size, SOUND_PCM );
	btnSoundOver.SetVolume( 30 );

	GuiImageData aboutData( about_png );
	GuiImageData aboutOverData( about_hover_png );
	GuiImageData gbaData( GBA_png );
	GuiImageData gbaOverData( GBA_hover_png );
	GuiImageData segaData( genesis_png );
	GuiImageData segaOverData( genesis_hover_png );
	GuiImageData n64Data( n64_png );
	GuiImageData n64OverData( n64_hover_png );
	GuiImageData nesData( NES_png );
	GuiImageData nesOverData( NES_hover_png );
	GuiImageData exitData( power_png );
	GuiImageData exitOverData( power_hover_png );
	GuiImageData snesUSData( SNES_US_png );
	GuiImageData snesUSOverData( SNES_US_hover_png );
	GuiImageData snesEUData( SNES_EU_png );
	GuiImageData snesEUOverData( SNES_EU_hover_png );

	GuiTrigger trigA;
	trigA.SetSimpleTrigger(-1, WPAD_BUTTON_A | WPAD_CLASSIC_BUTTON_A, PAD_BUTTON_A);
	GuiTrigger trigHome;
	trigHome.SetButtonOnlyTrigger(-1, WPAD_BUTTON_HOME | WPAD_CLASSIC_BUTTON_HOME, 0);

#ifdef DBG_LOCATIONS
	GuiTrigger trigb;
	trigb.SetButtonOnlyTrigger(-1, WPAD_BUTTON_B, 0);
	GuiTrigger trig1d;
	trig1d.SetButtonOnlyTrigger(-1, WPAD_BUTTON_1, 0);
	GuiTrigger trig2;
	trig2.SetButtonOnlyTrigger(-1, WPAD_BUTTON_2, 0);
	GuiTrigger trigPlus;
	trigPlus.SetButtonOnlyTrigger(-1, WPAD_BUTTON_PLUS, 0);
	GuiTrigger trigMinus;
	trigMinus.SetButtonOnlyTrigger(-1, WPAD_BUTTON_MINUS, 0);
	GuiTrigger trigL;
	trigL.SetButtonOnlyTrigger(-1, WPAD_BUTTON_LEFT, 0);
	GuiTrigger trigR;
	trigR.SetButtonOnlyTrigger(-1, WPAD_BUTTON_RIGHT, 0);
	GuiButton btnb( 0, 0 );
	btnb.SetTrigger( &trigb );
	GuiButton btn1( 0, 0 );
	btn1.SetTrigger( &trig1d );
	GuiButton btn2( 0, 0 );
	btn2.SetTrigger( &trig2 );
	GuiButton btnPlus( 0, 0 );
	btnPlus.SetTrigger( &trigPlus );
	GuiButton btnMinus( 0, 0 );
	btnMinus.SetTrigger( &trigMinus );
	GuiButton btnL( 0, 0 );
	btnL.SetTrigger( &trigL );
	GuiButton btnR( 0, 0 );
	btnR.SetTrigger( &trigR );

	GuiTrigger trig1;
	trig1.SetSimpleTrigger(-1, 0, 0 );
#else
	GuiTrigger trig1;
	trig1.SetSimpleTrigger(-1, WPAD_BUTTON_1 | WPAD_CLASSIC_BUTTON_Y, PAD_BUTTON_Y);

#endif

	//gratuitous brackets are just to allow folding the code in IDE to see shit better

	//message window.  make it into a non-clickable button just to manage it easier
	GuiText msgTxt("Trying to talk with the DVD drive...", 14, (GXColor){0, 0, 0, 255});
	GuiImage msgImg( 600, 24, (GXColor){ 255, 255, 255, 195 } );
	msgTxt.SetMaxWidth( msgImg.GetWidth() - 10 );
	GuiButton msgBtn( msgImg.GetWidth(), msgImg.GetHeight() );
	{
	msgBtn.SetAlignment( ALIGN_CENTRE, ALIGN_BOTTOM );
	msgBtn.SetPosition( 0, -16 );
	msgBtn.SetLabel( &msgTxt );
	msgBtn.SetImage( &msgImg );
	msgBtn.SetSelectable( false );
	msgBtn.SetState( STATE_DISABLED );
	}

	//gba button
	GuiImage gbaBtnImg( &gbaData );
	GuiImage gbaBtnImgOver( &gbaOverData );
	gbaBtnImg.SetAlignment( ALIGN_CENTRE, ALIGN_MIDDLE );
	gbaBtnImgOver.SetAlignment( ALIGN_CENTRE, ALIGN_MIDDLE );
	gbaBtnImg.SetWidescreenScale( widescreen );
	gbaBtnImgOver.SetWidescreenScale( widescreen );
	GuiButton gbaBtn( widescreen ? gbaData.GetWidth() * .80 : gbaData.GetWidth(), gbaData.GetHeight() );
	{
	gbaBtn.SetAlignment( ALIGN_LEFT, ALIGN_TOP );
	gbaBtn.SetPosition( widescreen ? 62 : 22, 21 );
	gbaBtn.SetImage( &gbaBtnImg );
	gbaBtn.SetImageOver( &gbaBtnImgOver );
	gbaBtn.SetSoundOver( &btnSoundOver );
	gbaBtn.SetTrigger( &trigA );
	gbaBtn.SetTrigger( &trig1 );
	}

	//nes button
	GuiImage nesBtnImg( &nesData );
	GuiImage nesBtnImgOver( &nesOverData );
	nesBtnImg.SetAlignment( ALIGN_CENTRE, ALIGN_MIDDLE );
	nesBtnImgOver.SetAlignment( ALIGN_CENTRE, ALIGN_MIDDLE );
	nesBtnImg.SetWidescreenScale( widescreen );
	nesBtnImgOver.SetWidescreenScale( widescreen );
	GuiButton nesBtn( widescreen ? nesData.GetWidth() * .80: nesData.GetWidth(), nesData.GetHeight() );
	{
	nesBtn.SetAlignment(ALIGN_LEFT, ALIGN_TOP);
	nesBtn.SetPosition( widescreen ? 84 : 42, 101 );
	nesBtn.SetImage( &nesBtnImg );
	nesBtn.SetImageOver( &nesBtnImgOver );
	nesBtn.SetSoundOver( &btnSoundOver );
	nesBtn.SetTrigger( &trigA );
	nesBtn.SetTrigger( &trig1 );
	}

	//snes button
	GuiImage snesBtnImg( &snesUSData );
	GuiImage snesBtnImgOver( &snesUSOverData );

	switch( region )
	{
	    case CONF_REGION_EU:
	    case CONF_REGION_JP:
		snesBtnImg.SetImage( &snesEUData );	//set different images for the button for these people
		snesBtnImgOver.SetImage( &snesEUOverData );
		break;
	    default:
		break;
	}
	snesBtnImg.SetAlignment( ALIGN_CENTRE, ALIGN_MIDDLE );
	snesBtnImgOver.SetAlignment( ALIGN_CENTRE, ALIGN_MIDDLE );
	snesBtnImg.SetWidescreenScale( widescreen );
	snesBtnImgOver.SetWidescreenScale( widescreen );
	GuiButton snesBtn( widescreen ? snesUSData.GetWidth() * .80 : snesUSData.GetWidth(), snesUSData.GetHeight() );
	{
	snesBtn.SetAlignment( ALIGN_LEFT, ALIGN_TOP );
	snesBtn.SetPosition( widescreen ? 99 : 63, 180 );
	snesBtn.SetImage( &snesBtnImg );
	snesBtn.SetImageOver( &snesBtnImgOver );
	snesBtn.SetSoundOver( &btnSoundOver );
	snesBtn.SetTrigger( &trigA );
	snesBtn.SetTrigger( &trig1 );
	}

	//wii64 button
	GuiImage wii64BtnImg( &n64Data );
	GuiImage wii64BtnImgOver( &n64OverData );	
	wii64BtnImg.SetAlignment( ALIGN_CENTRE, ALIGN_MIDDLE );
	wii64BtnImgOver.SetAlignment( ALIGN_CENTRE, ALIGN_MIDDLE );
	wii64BtnImg.SetWidescreenScale( widescreen );
	wii64BtnImgOver.SetWidescreenScale( widescreen );
	GuiButton wii64Btn( widescreen ? n64Data.GetWidth() * .80 : n64Data.GetWidth(), n64Data.GetHeight() );
	{
	wii64Btn.SetAlignment( ALIGN_LEFT, ALIGN_TOP );
	wii64Btn.SetPosition( widescreen ? 118 : 84, 258 );
	wii64Btn.SetImage( &wii64BtnImg );
	wii64Btn.SetImageOver( &wii64BtnImgOver );
	wii64Btn.SetSoundOver( &btnSoundOver );
	wii64Btn.SetTrigger( &trigA );
	wii64Btn.SetTrigger( &trig1 );
	}

	//sega button
	GuiImage segaBtnImg( &segaData );
	GuiImage segaBtnImgOver( &segaOverData );
	segaBtnImg.SetAlignment( ALIGN_CENTRE, ALIGN_MIDDLE );
	segaBtnImgOver.SetAlignment( ALIGN_CENTRE, ALIGN_MIDDLE );
	segaBtnImg.SetWidescreenScale( widescreen );
	segaBtnImgOver.SetWidescreenScale( widescreen );
	GuiButton segaBtn( widescreen ? segaData.GetWidth() * .80 : segaData.GetWidth(), segaData.GetHeight() );
	{
	segaBtn.SetAlignment( ALIGN_LEFT, ALIGN_TOP );
	segaBtn.SetPosition( widescreen ? 135 : 102, 336 );
	segaBtn.SetImage( &segaBtnImg );
	segaBtn.SetImageOver( &segaBtnImgOver );
	segaBtn.SetSoundOver( &btnSoundOver );
	segaBtn.SetTrigger( &trigA );
	segaBtn.SetTrigger( &trig1 );
	}

	GuiImage exitBtnImg( &exitData );
	GuiImage exitBtnImgOver( &exitOverData );
	exitBtnImg.SetAlignment( ALIGN_CENTRE, ALIGN_MIDDLE );
	exitBtnImgOver.SetAlignment( ALIGN_CENTRE, ALIGN_MIDDLE );
	exitBtnImg.SetWidescreenScale( widescreen );
	exitBtnImgOver.SetWidescreenScale( widescreen );
	GuiButton exitBtn( widescreen ? exitData.GetWidth() * .80 : exitData.GetWidth(), exitData.GetHeight() );
	{
	exitBtn.SetAlignment( ALIGN_LEFT, ALIGN_TOP );
	exitBtn.SetPosition( widescreen ? 437 : 450, 364 );
	exitBtn.SetImage(&exitBtnImg);
	exitBtn.SetImageOver(&exitBtnImgOver);
	exitBtn.SetSoundOver(&btnSoundOver);
	exitBtn.SetTrigger(&trigA);
	exitBtn.SetTrigger(&trigHome);
	}

	GuiImage aboutBtnImg( &aboutData );
	GuiImage aboutBtnImgOver( &aboutOverData );
	aboutBtnImg.SetAlignment( ALIGN_CENTRE, ALIGN_MIDDLE );
	aboutBtnImgOver.SetAlignment( ALIGN_CENTRE, ALIGN_MIDDLE );
	aboutBtnImg.SetWidescreenScale( widescreen );
	aboutBtnImgOver.SetWidescreenScale( widescreen );
	GuiButton aboutBtn( widescreen ? aboutData.GetWidth() * .80 : aboutData.GetWidth(), aboutData.GetHeight() );
	{
	aboutBtn.SetAlignment( ALIGN_LEFT, ALIGN_TOP );
	aboutBtn.SetPosition( widescreen ? 472 : 460, widescreen ? 309 : 300 );
	aboutBtn.SetImage(&aboutBtnImg);
	aboutBtn.SetImageOver(&aboutBtnImgOver);
	aboutBtn.SetSoundOver(&btnSoundOver);
	aboutBtn.SetTrigger(&trigA);
	}



	wii64Btn.SetEffect(EFFECT_SLIDE_LEFT | EFFECT_SLIDE_IN, 50);
	gbaBtn.SetEffect(EFFECT_SLIDE_LEFT | EFFECT_SLIDE_IN, 50);
	segaBtn.SetEffect(EFFECT_SLIDE_LEFT | EFFECT_SLIDE_IN, 50);
	snesBtn.SetEffect(EFFECT_SLIDE_LEFT | EFFECT_SLIDE_IN, 50);
	nesBtn.SetEffect(EFFECT_SLIDE_LEFT | EFFECT_SLIDE_IN, 50);

	exitBtn.SetEffect(EFFECT_SLIDE_RIGHT | EFFECT_SLIDE_IN, 50);
	aboutBtn.SetEffect(EFFECT_SLIDE_RIGHT | EFFECT_SLIDE_IN, 50);



#ifdef DBG_LOCATIONS

	int current = 0;
	int skip = 1;
	char dbgText[ 0x100 ];
	GuiElement * button = NULL;

	HaltGui();
	GuiWindow w(screenwidth, screenheight);
	w.Append( &gbaBtn );	    //0
	w.Append( &nesBtn );	    //1
	w.Append( &snesBtn );	    //2
	w.Append( &wii64Btn );	    //3
	w.Append( &segaBtn );	    //4

	w.Append( &msgBtn );	    //5
	w.Append( &exitBtn );	    //6
	w.Append( &aboutBtn );	    //7

	w.Append( &btnb );
	w.Append( &btn1 );
	w.Append( &btn2 );
	w.Append( &btnPlus );
	w.Append( &btnMinus );
	w.Append( &btnL );
	w.Append( &btnR );

	mainWindow->Append(&w);

	button = w.GetGuiElementAt( current );
	ResumeGui();

	while( menu == MENU_NONE )
	{
		usleep(THREAD_SLEEP);

		if( exitBtn.GetState() == STATE_CLICKED )
		{
		    menu = MENU_EXIT;
		}
		else if( aboutBtn.GetState() == STATE_CLICKED )
		{
		    aboutBtn.ResetState();
		}
		else if( btnb.GetState() == STATE_CLICKED )
		{
		    btnb.ResetState();
		    if( ++current > 7 )
		    //if( current++ >= w.GetSize() )
			current = 0;

		    button = w.GetGuiElementAt( current );
		    gprintf("[ %d ] %d, %d \n", current, button->GetPosX(), button->GetPosY());
		    sprintf( dbgText, "[ %d ] %d, %d", current, button->GetPosX(), button->GetPosY() );
		    msgTxt.SetText( dbgText );
		}
		else if( btn1.GetState() == STATE_CLICKED )
		{
		    btn1.ResetState();
		    button->SetPosition( button->GetPosX() - skip, button->GetPosY() );
		    gprintf("[ %d ] %d, %d \n", current, button->GetPosX(), button->GetPosY());
		    sprintf( dbgText, "[ %d ] %d, %d", current, button->GetPosX(), button->GetPosY() );
		    msgTxt.SetText( dbgText );
		}
		else if( btn2.GetState() == STATE_CLICKED )
		{
		    btn2.ResetState();
		    button->SetPosition( button->GetPosX() + skip, button->GetPosY() );
		    gprintf("[ %d ] %d, %d \n", current, button->GetPosX(), button->GetPosY());
		    sprintf( dbgText, "[ %d ] %d, %d", current, button->GetPosX(), button->GetPosY() );
		    msgTxt.SetText( dbgText );
		}
		else if( btnPlus.GetState() == STATE_CLICKED )
		{
		    btnPlus.ResetState();
		    button->SetPosition( button->GetPosX(), button->GetPosY() - skip );
		    gprintf("[ %d ] %d, %d \n", current, button->GetPosX(), button->GetPosY());
		    sprintf( dbgText, "[ %d ] %d, %d", current, button->GetPosX(), button->GetPosY() );
		    msgTxt.SetText( dbgText );
		}
		else if( btnMinus.GetState() == STATE_CLICKED )
		{
		    btnMinus.ResetState();
		    button->SetPosition( button->GetPosX() , button->GetPosY() + skip );
		    gprintf("[ %d ] %d, %d \n", current, button->GetPosX(), button->GetPosY());
		    sprintf( dbgText, "[ %d ] %d, %d", current, button->GetPosX(), button->GetPosY() );
		    msgTxt.SetText( dbgText );
		}
		else if( btnR.GetState() == STATE_CLICKED )
		{
		    btnR.ResetState();
		    skip++;
		    sprintf( dbgText, "Skip rate: %d", skip );
		    msgTxt.SetText( dbgText );
		}

		else if( btnL.GetState() == STATE_CLICKED )
		{
		    btnL.ResetState();
		    if( skip > 1 )skip--;
		    sprintf( dbgText, "Skip rate: %d", skip );
		    msgTxt.SetText( dbgText );
		}
	}

#else

	//HaltGui();
	GuiWindow w(screenwidth, screenheight);

	w.SetState( STATE_DISABLED );
	w.Append( &msgBtn );

	mainWindow->Append(&w);
	ResumeGui();



	//check that the expected dols are on the iso and disable buttons for missing ones
	// set the button as selected for the last emulator we loaded
	HaltGui();
	w.SetState( STATE_DEFAULT );
	w.ResetState();

	if( FSTIsMounted() )
	{
	    w.Append( &gbaBtn );//gui_window.cpp is changed now to cycle through the buttons in the order they are appended
	    w.Append( &nesBtn );//so order is important here
	    w.Append( &snesBtn );
	    w.Append( &wii64Btn );
	    w.Append( &segaBtn );
	    w.Append( &aboutBtn );
	}
	w.Append( &exitBtn );//somthing went wrong mounting the filesystem.  just give them the exit button

	ResumeGui();

	if( FSTIsMounted() )
	{

	    //wait for all the buttons to be on screen
	    while( snesBtn.GetEffect() > 0 )
		usleep( THREAD_SLEEP );

	    //select the button that goes to the last emulator we started
	    if( !strncmp( lastEmu, SNES_PATH, strlen( SNES_PATH ) ) )
		w.SelectElement( &snesBtn );

	    else if( !strncmp( lastEmu, GBA_PATH, strlen( GBA_PATH ) ) )
		w.SelectElement( &gbaBtn );

	    else if( !strncmp( lastEmu, N64_PATH, strlen( N64_PATH ) ) )
		w.SelectElement( &wii64Btn );

	    else if( !strncmp( lastEmu, SEGA_PATH, strlen( SEGA_PATH ) ) )
		w.SelectElement( &segaBtn );

	    else if( !strncmp( lastEmu, NES_PATH, strlen( NES_PATH ) ) )
		w.SelectElement( &nesBtn );

	    else w.SelectElement( &aboutBtn );

	    //clear this, we're done with it
	    lastEmu[ 0 ] = 0;

	    //disable buttons for emus that are missing from the iso.
	    if( !CheckFile( N64_PATH ) )
		wii64Btn.SetState( STATE_DISABLED );

	    if( !CheckFile( NES_PATH ) )
		nesBtn.SetState( STATE_DISABLED );

	    if( !CheckFile( GBA_PATH ) )
		gbaBtn.SetState( STATE_DISABLED );

	    if( !CheckFile( SNES_PATH ) )
		snesBtn.SetState( STATE_DISABLED );

	    if( !CheckFile( SEGA_PATH ) )
		segaBtn.SetState( STATE_DISABLED );

	    msgTxt.SetText( "Disc opening and read.  Pick an emulator.");
	}
	else//what error do we display
	{
	    if( WDVD_GetHandle() < 0 )
		msgTxt.SetText( "Error strarting the DI driver!");//something really bad happened here
	    else if( !WDVD_IsPartitionOpen() )
		msgTxt.SetText( "Error opening the disc partition!");//they arent using an IOS that allows fakesigning
	    else msgTxt.SetText( "Error reading the disc filesystem!");//they botched something up making the iso
	}


	while( menu == MENU_NONE )
	{
		usleep(THREAD_SLEEP);

		if( wii64Btn.GetState() == STATE_CLICKED )
		{

		    msgTxt.SetText( "Loading N64...");
		    int r = wii64Btn.GetTriggeredTrigger();
		    w.SetState( STATE_DISABLED );
		    if( !LoadHomebrew( N64_PATH ) )
		    {
			wii64Btn.SetState( STATE_DEFAULT );
			w.SetState( STATE_DEFAULT );
			msgTxt.SetText( "Error reading " N64_PATH );
		    }

		    else
		    {
			if( r )
			    AddBootArgument( N64_USB_PATH );

			else
			    AddBootArgument( N64_SD_PATH );

			menu = MENU_EXIT;
		    }


		}
		else if( gbaBtn.GetState() == STATE_CLICKED )
		{
		    msgTxt.SetText( "Loading VBA-GX...");
		    int r = gbaBtn.GetTriggeredTrigger();
		    w.SetState( STATE_DISABLED );
		    if( !LoadHomebrew( GBA_PATH ) )
		    {
			gbaBtn.SetState( STATE_DEFAULT );
			w.SetState( STATE_DEFAULT );
			msgTxt.SetText( "Error reading " GBA_PATH );
		    }

		    else
		    {
			if( r )
			    AddBootArgument( GBA_USB_PATH );

			else
			    AddBootArgument( GBA_SD_PATH );

			menu = MENU_EXIT;
		    }
		}
		else if( snesBtn.GetState() == STATE_CLICKED )
		{
		    msgTxt.SetText( "Loading SNES9X-GX...");
		    int r = snesBtn.GetTriggeredTrigger();
		    w.SetState( STATE_DISABLED );
		    if( !LoadHomebrew( SNES_PATH ) )
		    {
			snesBtn.SetState( STATE_DEFAULT );
			w.SetState( STATE_DEFAULT );
			msgTxt.SetText( "Error reading " SNES_PATH );
		    }

		    else
		    {
			if( r )
			    AddBootArgument( SNES_USB_PATH );

			else
			    AddBootArgument( SNES_SD_PATH );

			menu = MENU_EXIT;
		    }
		}
		else if( nesBtn.GetState() == STATE_CLICKED )
		{
		    msgTxt.SetText( "Loading FCEU-GX...");
		    int r = nesBtn.GetTriggeredTrigger();
		    w.SetState( STATE_DISABLED );
		    if( !LoadHomebrew( NES_PATH ) )
		    {
			nesBtn.SetState( STATE_DEFAULT );
			w.SetState( STATE_DEFAULT );
			msgTxt.SetText( "Error reading " NES_PATH );
		    }

		    else
		    {
			if( r )
			    AddBootArgument( NES_USB_PATH );

			else
			    AddBootArgument( NES_SD_PATH );

			menu = MENU_EXIT;
		    }

		}
		else if( segaBtn.GetState() == STATE_CLICKED )
		{
		    if( !HaveSD() )
		    {
			msgTxt.SetText( "Insert an SD card to use this emulator");
			segaBtn.SetState( STATE_DEFAULT );
			w.SetState( STATE_DEFAULT );
			continue;
		    }
		    msgTxt.SetText( "Loading GenPlus-GX...");
		    //int r = segaBtn.GetTriggeredTrigger();
		    w.SetState( STATE_DISABLED );
		    if( !LoadHomebrew( SEGA_PATH ) )
		    {
			segaBtn.SetState( STATE_DEFAULT );
			w.SetState( STATE_DEFAULT );
			msgTxt.SetText( "Error reading " SEGA_PATH );
		    }


		    else
		    {	//gensplus doesn't like being started without an SD card
			//if( r )
			//    AddBootArgument( SEGA_USB_PATH );

			//else
			    AddBootArgument( SEGA_SD_PATH );
			    if( firstRun )//show the "legal" screen only the first time this emulator is started
				AddBootArgument( "legal" );

			menu = MENU_EXIT;
		    }
		}
		else if( exitBtn.GetState() == STATE_CLICKED )
		{
			menu = MENU_EXIT;
		}
		else if( aboutBtn.GetState() == STATE_CLICKED )
		{
			menu = MENU_ABOUT;
		}
		else if( ResetRequested )
		{
		    msgTxt.SetText( "Resetting...");
		    w.SetState( STATE_DISABLED );
		    if( WDVD_GetMainDol( (void*)homebrewbuffer, &homebrewsize ) < 0 )
		    {
			SYS_ResetSystem(SYS_RETURNTOMENU, 0, 0);
		    }

		    else
		    {
			AddBootArgument( "reset" );
			menu = MENU_EXIT;
		    }
		}
	}

	wii64Btn.SetEffect(EFFECT_SLIDE_LEFT | EFFECT_SLIDE_OUT, 50);
	gbaBtn.SetEffect(EFFECT_SLIDE_LEFT | EFFECT_SLIDE_OUT, 50);
	segaBtn.SetEffect(EFFECT_SLIDE_LEFT | EFFECT_SLIDE_OUT, 50);
	snesBtn.SetEffect(EFFECT_SLIDE_LEFT | EFFECT_SLIDE_OUT, 50);
	nesBtn.SetEffect(EFFECT_SLIDE_LEFT | EFFECT_SLIDE_OUT, 50);

	exitBtn.SetEffect(EFFECT_SLIDE_RIGHT | EFFECT_SLIDE_OUT, 50);
	aboutBtn.SetEffect(EFFECT_SLIDE_RIGHT | EFFECT_SLIDE_OUT, 50);

	while( FSTIsMounted() && snesBtn.GetEffect() > 0 )
	    usleep( THREAD_SLEEP );

#endif
	HaltGui();
	mainWindow->Remove( &w );
	//ResumeGui();
	return menu;
}

static int MenuAbout()
{
    int menu = MENU_EMU_SELECT;
	bool exit = false;
	int i = 0;
	int y = 20;

	GuiWindow creditsWindow( screenwidth,screenheight );
	GuiWindow creditsWindowBox(580,448);
	creditsWindowBox.SetAlignment(ALIGN_CENTRE, ALIGN_MIDDLE);


	GuiImageData creditsBox( about_BG_png );
	GuiImage creditsBoxImg( &creditsBox );
	creditsBoxImg.SetAlignment( ALIGN_CENTRE, ALIGN_MIDDLE );
	creditsBoxImg.SetScale( 620, 460 );
	creditsBoxImg.SetPosition( 0, 0 );
	creditsWindow.Append(&creditsBoxImg);

	int numEntries = 27;
	int textSpacing = 18;
	GuiText * txt[numEntries];

	txt[i] = new GuiText("- PunEmu -", 22, (GXColor){0, 0, 0, 255});
	txt[i]->SetAlignment(ALIGN_CENTRE, ALIGN_TOP); txt[i]->SetPosition(0,y); i++; y+=28;

	txt[i] = new GuiText("This dol is part of an image by giantpune.", 18, (GXColor){0, 0, 0, 255});
	txt[i]->SetAlignment(ALIGN_CENTRE, ALIGN_TOP); txt[i]->SetPosition(0,y); i++; y+=30;

	txt[i]->SetPresets(14, (GXColor){0, 0, 0, 255}, 0,
			FTGX_JUSTIFY_LEFT | FTGX_ALIGN_TOP, ALIGN_LEFT, ALIGN_TOP);

	txt[i] = new GuiText("Coding, Composing");
	txt[i]->SetPosition(60,y); i++;
	txt[i] = new GuiText("giantpune");
	txt[i]->SetPosition( 350, y ); i++; y+=textSpacing;

	txt[i] = new GuiText("Graphics");
	txt[i]->SetPosition(60,y); i++;
	txt[i] = new GuiText("NeoRame");
	txt[i]->SetPosition( 350, y ); i++; y+=textSpacing;
	txt[i] = new GuiText("Banner");
	txt[i]->SetPosition(60,y); i++;
	txt[i] = new GuiText("WiiThemeTeam");
	txt[i]->SetPosition(350,y); i++; y+=textSpacing + 20;

	txt[i] = new GuiText("The image was built with QtWitGui and Wiimms\' Tools", 18, (GXColor){0, 0, 0, 255});
	txt[i]->SetAlignment(ALIGN_CENTRE, ALIGN_TOP); txt[ i ]->SetPosition(0,y); i++; y+=30;

	txt[i] = new GuiText("Wiimms\' Tools");
	txt[i]->SetPosition(60,y); i++;
	txt[i] = new GuiText("http://wit.wiimm.de/");
	txt[i]->SetPosition(220,y); i++; y+=textSpacing;

	txt[i] = new GuiText("QtWitGui");
	txt[i]->SetPosition(60,y); i++;
	txt[i] = new GuiText("http://code.google.com/p/qtwitgui/");
	txt[i]->SetPosition(220,y); i++; y+=textSpacing + 20;

	txt[i] = new GuiText("Emulators modified from source from various others.", 18, (GXColor){0, 0, 0, 255});
	txt[i]->SetAlignment(ALIGN_CENTRE, ALIGN_TOP); txt[i]->SetPosition(0,y); i++; y+=30;

	txt[i] = new GuiText("FCE Ultra GX");
	txt[i]->SetPosition(60,y); i++;
	txt[i] = new GuiText("http://code.google.com/p/fceugc/");
	txt[i]->SetPosition(220,y); i++; y+=textSpacing;

	txt[i] = new GuiText("VBA-GX");
	txt[i]->SetPosition(60,y); i++;
	txt[i] = new GuiText("http://code.google.com/p/vba-wii/");
	txt[i]->SetPosition(220,y); i++; y+=textSpacing;

	txt[i] = new GuiText("Snes9x-GX");
	txt[i]->SetPosition(60,y); i++;
	txt[i] = new GuiText("http://code.google.com/p/snes9x-gx/");
	txt[i]->SetPosition(220,y); i++; y+=textSpacing;

	txt[i] = new GuiText("GensPlus GX");
	txt[i]->SetPosition(60,y); i++;
	txt[i] = new GuiText("http://code.google.com/p/genplus-gx/");
	txt[i]->SetPosition(220,y); i++; y+=textSpacing;

	txt[i] = new GuiText("Wii64");
	txt[i]->SetPosition(60,y); i++;
	txt[i] = new GuiText("http://emulatemii.com/");
	txt[i]->SetPosition(220,y); i++; y+=38;


	txt[i]->SetPresets(14, (GXColor){0, 0, 0, 255}, 0,
		FTGX_JUSTIFY_CENTER | FTGX_ALIGN_TOP, ALIGN_CENTRE, ALIGN_TOP);

	txt[i] = new GuiText("This software is open source and may be copied,");
	txt[i]->SetPosition(0,y); i++; y+=20;
	txt[i] = new GuiText("distributed, or modified under the terms of the");
	txt[i]->SetPosition(0,y); i++; y+=20;
	txt[i] = new GuiText("GNU General Public License (GPL) Version 2.");
	txt[i]->SetPosition(0,y); i++; y+=20;

	for(i=0; i < numEntries; i++)
		creditsWindowBox.Append(txt[i]);

	//HaltGui();
	creditsWindow.Append(&creditsWindowBox);
	mainWindow->Append( &creditsWindow );
	ResumeGui();

	while( !exit && !ResetRequested )
	{
		if((userInput[0].wpad->btns_d || userInput[0].pad.btns_d) ||
		   (userInput[1].wpad->btns_d || userInput[1].pad.btns_d) ||
		   (userInput[2].wpad->btns_d || userInput[2].pad.btns_d) ||
		   (userInput[3].wpad->btns_d || userInput[3].pad.btns_d))
		{
			exit = true;
		}
		usleep(THREAD_SLEEP);
	}

	HaltGui();

	mainWindow->Remove( &creditsWindow );
	// clear buttons pressed
	for( i=0; i < 4; i++ )
	{
		userInput[ i ].wpad->btns_d = 0;
		userInput[ i ].pad.btns_d = 0;
	}

	for( i=0; i < numEntries; i++ )
		delete txt[ i ];


	//ResumeGui();

	return menu;
}

int MenuStrapScreen( int silent )
{
    GuiImageData strapData( safety_EN_png );	    //nice professional looking safety screens
    GuiImageData strapData169( safety_EN_16_9_png );//nintendo would be proud

    //only show the safety screen on first run/reset
    if( silent != MENU_STRAP_SILENT )
    {
	Menu_DrawImg( 0, 0, 640, 480, widescreen ? strapData169.GetImage() : strapData.GetImage(), 0, 1, 1, 255 );
	Menu_Render();
    }

    if( !WDVD_Init() )
    {
	if( !WDVD_OpenDataPartition() )
	{
	    FST_Mount();
	}
    }

    if( silent != MENU_STRAP_SILENT )
    {
	sleep( 4 );
	//fade out strap screen
	for( int i = 0; i <= 255; i += 5 )
	{
		Menu_DrawImg( 0, 0, 640, 480, widescreen ? strapData169.GetImage() : strapData.GetImage(), 0, 1, 1, 255 );
		Menu_DrawRectangle( 0, 0, screenwidth, screenheight, (GXColor){ 0, 0, 0, i }, 1 );
		Menu_Render();
	}
	firstRun = true;
    }

    return MENU_EMU_SELECT;
}

/****************************************************************************
 * MainMenu
 ***************************************************************************/
void MainMenu(int menu)
{
	int currentMenu = menu;

	mainWindow = new GuiWindow(screenwidth, screenheight);

	GuiImageData bgData( background_png );
	bgImg = new GuiImage( &bgData );

	//if( menu == MENU_STRAP )
	currentMenu = MenuStrapScreen( menu );

	bgMusic = new GuiSound( bg_music_ogg, bg_music_ogg_size, SOUND_OGG );
	bgMusic->SetLoop( true );
	bgMusic->SetVolume( 50 );
	bgMusic->Play(); // startup music


	mainWindow->Append( bgImg );

	//fade in from black to "disabled"
	for( int i = 255; i >= 0x80; i -= 3 )
	{
		Menu_DrawImg( 0, 0, 640, 480, bgData.GetImage(), 0, 1, 1, 255 );
		Menu_DrawRectangle( 0, 0, screenwidth, screenheight, (GXColor){ 0, 0, 0, i },1 );
		Menu_Render();
	}

	pointer[0] = new GuiImageData(player1_point_png);
	pointer[1] = new GuiImageData(player2_point_png);
	pointer[2] = new GuiImageData(player3_point_png);
	pointer[3] = new GuiImageData(player4_point_png);



	//ResumeGui();

	while(currentMenu != MENU_EXIT)
	{
		switch (currentMenu)
		{
			case MENU_ABOUT:
				currentMenu = MenuAbout();
				break;
			case MENU_EMU_SELECT:
			default: // unrecognized menu
				currentMenu = MenuEmuSelector();
				break;
		}
	}

	if( WDVD_IsPartitionOpen() )
	    WDVD_ClosePartition();

	ResumeGui();
	ExitRequested = 1;
	//while(1) usleep(THREAD_SLEEP);

	HaltGui();



	bgMusic->Stop();
	delete bgMusic;
	delete bgImg;
	delete mainWindow;

	delete pointer[0];
	delete pointer[1];
	delete pointer[2];
	delete pointer[3];

	mainWindow = NULL;
	while(1) usleep(THREAD_SLEEP);
}
