/**
 * Wii64 - fileBrowser-DVD.c
 * Copyright (C) 2007, 2008, 2009 emu_kidid
 * 
 * fileBrowser module for ISO9660 DVD Discs
 *
 * Wii64 homepage: http://www.emulatemii.com
 * email address:  emukidid@gmail.com
 *
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
**/


#include <string.h>
#include <unistd.h>
#include <malloc.h>
#include <ogc/dvd.h>
#include <ogc/machine/processor.h>
#include "fileBrowser.h"
#include "../main/gc_dvd.h"
#ifdef HW_RVL
#include <di/di.h>
#endif

/* DVD Globals */
int dvd_init = 0;

/* Worked out manually from my original Disc */
#define ZELDA_BONUS_ID "D43U01"
#define ZELDA_CLCTR_ID "PZLP01"
#define ZELDA_BONUS_OOT 0x54FBEEF4ULL
#define ZELDA_BONUS_MQ  0x52CCC5FCULL
#define ZELDA_CLCTR_OOT 0x3B9D1FC0ULL
#define ZELDA_CLCTR_MM  0x0C4E1FC0ULL
#define ZELDA_SIZE      0x2000000
#define ZELDA_OOT_NAME  "Zelda - Ocarina of Time"
#define ZELDA_MQ_NAME   "Zelda - Ocarina of Time Master Quest"
#define ZELDA_MM_NAME   "Zelda - Majoras Mask"

fileBrowser_file topLevel_DVD =
	{ "\\", // file name
	  0ULL,      // discoffset (u64)
	  0,         // offset
	  0,         // size
	  FILE_BROWSER_ATTR_DIR
	};

typedef struct {
    u8 filetype;
    char name_offset[3];
    u32 fileoffset;
    u32 filelen;
} __attribute__((packed)) FST_ENTRY;

int fileBrowser_DVD_readDir(fileBrowser_file* ffile, fileBrowser_file** dir){	
  
    u64 discOff = ffile->discoffset;
//    gprintf("fileBrowser_DVD_readDir( %016llx )\n", discOff );


    int num_entries = 0;//, ret = 0;

    if( !IsFSTLoaded() )
    {
	gprintf("\tfst not loaded!\n");
	return 0;
    }

    //the whole fst
    FST_ENTRY *fst = (FST_ENTRY *)fstbuffer;
    u32 name_table_offset = fst->filelen * 0xC;
    char *name_table = (char *)( fstbuffer + name_table_offset );

    //find the entry that matches ffile
    FST_ENTRY *currentDir = (FST_ENTRY *)( fstbuffer + ( discOff * 0xC ));
    u32 name_offset = *((u32 *)currentDir) & 0x00ffffff;
//    gprintf("\tlisting: %s\n", name_table + name_offset );
    if( !currentDir->filetype )
    {
	gprintf("ERROR!!  %s is not a directory\n", name_table + name_offset );
	return 0;
    }

    int i = discOff + 1;
    u32 cnt = 0;
    //count entries in this directory
    while( i < currentDir->filelen )
    {
	FST_ENTRY *entry = (FST_ENTRY *)( fstbuffer + ( i * 0xC ));
	if( entry->filetype )
	    i += ( entry->filelen - i );
	else
	    i++;

	cnt++;
    }
    //add 1 for ".." but not on the root entry
    if( discOff )
	cnt++;

    //allocate memory
    *dir = malloc( cnt * sizeof(fileBrowser_file) );
    if( !*dir )
	return 0;

    i = discOff + 1;
    cnt = 0;
    if( ffile->discoffset )
    {
	strcpy( (*dir)[ cnt ].name, ".." );
	(*dir)[ cnt ].discoffset = currentDir->fileoffset;
	(*dir)[ cnt ].offset = 0;
	(*dir)[ cnt ].size   = ffile->size;
	(*dir)[ cnt ].attr   = FILE_BROWSER_ATTR_DIR;
	cnt++;
    }

    //now add entries to dir
    while( i < currentDir->filelen )
    {
	FST_ENTRY *entry = (FST_ENTRY *)( fstbuffer + ( i * 0xC ));
	u32 name_offset = *((u32 *)entry) & 0x00ffffff;
	//gprintf("\n[ %d ] [ %u ]%s\n", i, entry->filelen, name_table + name_offset );

	strcpy( (*dir)[ cnt ].name, name_table + name_offset );

	(*dir)[ cnt ].offset = 0;
	(*dir)[ cnt ].size   = entry->filelen;
	(*dir)[ cnt ].attr	 = 0;



	if( entry->filetype )//folder
	{
	    (*dir)[ cnt ].discoffset = i;
	    (*dir)[ cnt ].attr   = FILE_BROWSER_ATTR_DIR;
	    i += ( entry->filelen - i );
	}
	else//file
	{
	    (*dir)[ cnt ].discoffset = (u64)entry->fileoffset << 2LL;
	    i++;
	}

	cnt++;
    }

    //gprintf("\tdirectory has %u entries\n", cnt );
    return cnt;
}

int fileBrowser_DVD_seekFile(fileBrowser_file* file, unsigned int where, unsigned int type){
	if(type == FILE_BROWSER_SEEK_SET) file->offset = where;
	else if(type == FILE_BROWSER_SEEK_CUR) file->offset += where;
	else file->offset = file->size + where;
	return 0;
}

int fileBrowser_DVD_readFile(fileBrowser_file* file, void* buffer, unsigned int length)
{
    if( WDVD_Read( buffer, length, file->discoffset+file->offset ) < 0 )
	return 0;

    file->offset += length;
    return length;
	/*int bytesread = read_safe(buffer,file->discoffset+file->offset,length);
	if(bytesread > 0)
		file->offset += bytesread;
	return bytesread;*/
}

int fileBrowser_DVD_init(fileBrowser_file* file){

    if( !IsFSTLoaded() )
	return LoadFST() >= 0;

    return 0;
}

int fileBrowser_DVD_deinit(fileBrowser_file* file) {

    ForgetFST();
    return 0;
}

