/**
 * Wii64/Cube64 - gc_dvd.c
 * Copyright (C) 2007, 2008, 2009, 2010 emu_kidid
 * 
 * DVD Reading support for GC/Wii
 *
 * Wii64 homepage: http://www.emulatemii.com
 * email address: emukidid@gmail.com
 *
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
**/

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <ogc/dvd.h>
#include <malloc.h>
#include <string.h>
#include <gccore.h>
#include <unistd.h>
#include "gc_dvd.h"
//#ifdef WII
//#include <di/di.h>
//#endif

/* DVD Stuff */
u32 dvd_hard_init = 0;
static u32 read_cmd = NORMAL;
static int last_current_dir = -1;
int is_unicode,files;
file_entries *DVDToc = NULL; //Dynamically allocate this

typedef struct {
    u32 dol_offset;
    u32 fst_offset;
    u32 fst_size;
    u32 fst_size2;
} __attribute__ ((packed)) FST_INFO;

u8 *fstbuffer = NULL;
static bool fstLoaded = false;

#ifdef HW_DOL
#define mfpvr()   ({unsigned int rval; \
      asm volatile("mfpvr %0" : "=r" (rval)); rval;})
#endif
#ifdef HW_RVL
volatile unsigned long* dvd = (volatile unsigned long*)0xCD806000;
#else
volatile unsigned long* dvd = (volatile unsigned long*)0xCC006000;
#endif

int have_hw_access() {
  if((*(volatile unsigned int*)HW_ARMIRQMASK)&&(*(volatile unsigned int*)HW_ARMIRQFLAG)) {
    return 1;
  }
  return 0;
}

int init_dvd() {
// Gamecube Mode
#ifdef HW_DOL
  if(mfpvr()!=GC_CPU_VERSION) //GC mode on Wii, modchip required
  {
    DVD_Reset(DVD_RESETHARD);
    dvd_read_id();
    if(!dvd_get_error()) {
      return 0; //we're ok
    }
  }
  else      //GC, no modchip even required :)
  {
    DVD_Reset(DVD_RESETHARD);
    DVD_Mount ();
    if(!dvd_get_error()) {
      return 0; //we're ok
    }
  }
  if(dvd_get_error()>>24) {
    return NO_DISC;
  }
  return -1;

#endif
// Wii (Wii mode)
#ifdef HW_RVL

  //TODO
  /*if(!have_hw_access()) {
    return NO_HW_ACCESS;
  }
  if((dvd_get_error()>>24) == 1) {
    return NO_DISC;
  }
  
  if((!dvd_hard_init) || (dvd_get_error())) {
    DI_Mount();
    while(DI_GetStatus() & DVD_INIT) usleep(20000);
    dvd_hard_init=1;
  }

  if((dvd_get_error()&0xFFFFFF)==0x053000) {
    read_cmd = DVDR;
  }
  else {
    read_cmd = NORMAL;
  }*/
  return 0;
#endif
}

int dvd_read_id()
{
#ifdef HW_RVL
  DVD_LowRead64((void*)0x80000000, 32, 0ULL);  //for easter egg disc support
  return 0;
#endif
	dvd[0] = 0x2E;
	dvd[1] = 0;
	dvd[2] = 0xA8000040;
	dvd[3] = 0;
	dvd[4] = 0x20;
	dvd[5] = 0x80000000;
	dvd[6] = 0x20;
	dvd[7] = 3; // enable reading!
	while (dvd[7] & 1);
	if (dvd[0] & 0x4)
		return 1;
	return 0;
}

unsigned int dvd_get_error(void)
{
	dvd[2] = 0xE0000000;
	dvd[8] = 0;
	dvd[7] = 1; // IMM
	while (dvd[7] & 1);
	return dvd[8];
}


void dvd_motor_off()
{
	dvd[0] = 0x2E;
	dvd[1] = 0;
	dvd[2] = 0xe3000000;
	dvd[3] = 0;
	dvd[4] = 0;
	dvd[5] = 0;
	dvd[6] = 0;
	dvd[7] = 1; // IMM
	while (dvd[7] & 1);
}

/* 
DVD_LowRead64(void* dst, unsigned int len, uint64_t offset)
  Read Raw, needs to be on sector boundaries
  Has 8,796,093,020,160 byte limit (8 TeraBytes)
  Synchronous function.
    return -1 if offset is out of range
*/
int DVD_LowRead64(void* dst, unsigned int len, uint64_t offset)
{
  if(offset>>2 > 0xFFFFFFFF)
    return -1;
    
  if ((((int)dst) & 0xC0000000) == 0x80000000) // cached?
		dvd[0] = 0x2E;
	dvd[1] = 0;
	dvd[2] = read_cmd;
	dvd[3] = read_cmd == DVDR ? offset>>11 : offset >> 2;
	dvd[4] = read_cmd == DVDR ? len>>11 : len;
	dvd[5] = (unsigned long)dst;
	dvd[6] = len;
	dvd[7] = 3; // enable reading!
	DCInvalidateRange(dst, len);
	while (dvd[7] & 1);

	if (dvd[0] & 0x4)
		return 1;
	return 0;
}

static char error_str[256];
char *dvd_error_str()
{
  unsigned int err = dvd_get_error();
  if(!err) return "OK";
  
  memset(&error_str[0],0,256);
  switch(err>>24) {
    case 0:
      break;
    case 1:
      strcpy(&error_str[0],"Lid open");
      break;
    case 2:
      strcpy(&error_str[0],"No disk/Disk changed");
      break;
    case 3:
      strcpy(&error_str[0],"No disk");
      break;  
    case 4:
      strcpy(&error_str[0],"Motor off");
      break;  
    case 5:
      strcpy(&error_str[0],"Disk not initialized");
      break;
  }
  switch(err&0xFFFFFF) {
    case 0:
      break;
    case 0x020400:
      strcat(&error_str[0]," Motor Stopped");
      break;
    case 0x020401:
      strcat(&error_str[0]," Disk ID not read");
      break;
    case 0x023A00:
      strcat(&error_str[0]," Medium not present / Cover opened");
      break;  
    case 0x030200:
      strcat(&error_str[0]," No Seek complete");
      break;  
    case 0x031100:
      strcat(&error_str[0]," UnRecoverd read error");
      break;
    case 0x040800:
      strcat(&error_str[0]," Transfer protocol error");
      break;
    case 0x052000:
      strcat(&error_str[0]," Invalid command operation code");
      break;
    case 0x052001:
      strcat(&error_str[0]," Audio Buffer not set");
      break;  
    case 0x052100:
      strcat(&error_str[0]," Logical block address out of range");
      break;  
    case 0x052400:
      strcat(&error_str[0]," Invalid Field in command packet");
      break;
    case 0x052401:
      strcat(&error_str[0]," Invalid audio command");
      break;
    case 0x052402:
      strcat(&error_str[0]," Configuration out of permitted period");
      break;
    case 0x053000:
      strcat(&error_str[0]," DVD-R"); //?
      break;
    case 0x053100:
      strcat(&error_str[0]," Wrong Read Type"); //?
      break;
    case 0x056300:
      strcat(&error_str[0]," End of user area encountered on this track");
      break;  
    case 0x062800:
      strcat(&error_str[0]," Medium may have changed");
      break;  
    case 0x0B5A01:
      strcat(&error_str[0]," Operator medium removal request");
      break;
  }
  if(!error_str[0])
    sprintf(&error_str[0],"Unknown %08X",err);
  return &error_str[0];
  
}


int read_sector(void* buffer, uint32_t sector)
{
	return DVD_LowRead64(buffer, 2048, sector * 2048);
}

int read_safe(void* dst, uint64_t offset, int len)
{
	int ol = len;
	int ret = 0;	
  unsigned char* sector_buffer = (unsigned char*)memalign(32,32768);
	while (len)
	{
		ret |= DVD_LowRead64(sector_buffer, 32768, offset);
		uint32_t off = offset & 32767;

		int rl = 32768 - off;
		if (rl > len)
			rl = len;
		else 
		  rl = 32768;
		memcpy(dst, sector_buffer, rl);	

		offset += rl;
		len -= rl;
		dst += rl;
	}
	free(sector_buffer);
	if(ret)
  	return -1;
  	
  if(dvd_get_error())
    init_dvd();

	return ol;
}

int read_direntry(unsigned char* direntry)
{
       int nrb = *direntry++;
       ++direntry;

       int sector;

       direntry += 4;
       sector = (*direntry++) << 24;
       sector |= (*direntry++) << 16;
       sector |= (*direntry++) << 8;
       sector |= (*direntry++);        

       int size;

       direntry += 4;

       size = (*direntry++) << 24;
       size |= (*direntry++) << 16;
       size |= (*direntry++) << 8;
       size |= (*direntry++);

       direntry += 7; // skip date

       int flags = *direntry++;
       ++direntry; ++direntry; direntry += 4;

       int nl = *direntry++;

       char* name = DVDToc->file[files].name;

       DVDToc->file[files].sector = sector;
       DVDToc->file[files].size = size;
       DVDToc->file[files].flags = flags;

       if ((nl == 1) && (direntry[0] == 1)) // ".."
       {
               DVDToc->file[files].name[0] = 0;
               if (last_current_dir != sector)
                       files++;
       }
       else if ((nl == 1) && (direntry[0] == 0))
       {
               last_current_dir = sector;
       }
       else
       {
               if (is_unicode)
               {
                       int i;
                       for (i = 0; i < (nl / 2); ++i)
                               name[i] = direntry[i * 2 + 1];
                       name[i] = 0;
                       nl = i;
               }
               else
               {
                       memcpy(name, direntry, nl);
                       name[nl] = 0;
               }

               if (!(flags & 2))
               {
                       if (name[nl - 2] == ';')
                               name[nl - 2] = 0;

                       int i = nl;
                       while (i >= 0)
                               if (name[i] == '.')
                                       break;
                               else
                                       --i;

                       ++i;

               }
               else
               {
                       name[nl++] = '/';
                       name[nl] = 0;
               }

               files++;
       }

       return nrb;
}

void read_directory(int sector, int len)
{
  int ptr = 0;
  unsigned char *sector_buffer = (unsigned char*)memalign(32,2048);
  read_sector(sector_buffer, sector);
  
  files = 0;
  memset(DVDToc,0,sizeof(file_entries));
  while (len > 0)
  {
    ptr += read_direntry(sector_buffer + ptr);
    if (ptr >= 2048 || !sector_buffer[ptr])
    {
      len -= 2048;
      sector++;
      read_sector(sector_buffer, sector);
      ptr = 0;
    }
  }
  free(sector_buffer);
}

int dvd_read_directoryentries(uint64_t offset, int size) {
  int sector = 16;
  unsigned char *bufferDVD = (unsigned char*)memalign(32,2048);
  struct pvd_s* pvd = 0;
  struct pvd_s* svd = 0;
  
  if(DVDToc)
  {
    free(DVDToc);
    DVDToc = NULL;
  }
  DVDToc = memalign(32,sizeof(file_entries));
  
  while (sector < 32)
  {
    if (read_sector(bufferDVD, sector))
    {
      free(bufferDVD);
      free(DVDToc);
      DVDToc = NULL;
      return FATAL_ERROR;
    }
    if (!memcmp(((struct pvd_s *)bufferDVD)->id, "\2CD001\1", 8))
    {
      svd = (void*)bufferDVD;
      break;
    }
    ++sector;
  }
  
  
  if (!svd)
  {
    sector = 16;
    while (sector < 32)
    {
      if (read_sector(bufferDVD, sector))
      {
        free(bufferDVD);
        free(DVDToc);
        DVDToc = NULL;
        return FATAL_ERROR;
      }
      
      if (!memcmp(((struct pvd_s *)bufferDVD)->id, "\1CD001\1", 8))
      {
        pvd = (void*)bufferDVD;
        break;
      }
      ++sector;
    }
  }
  
  if ((!pvd) && (!svd))
  {
    free(bufferDVD);
    free(DVDToc);
    DVDToc = NULL;
    return NO_ISO9660_DISC;
  }
  
  files = 0;
  if (svd)
  {
    is_unicode = 1;
    read_direntry(svd->root_direntry);
  }
  else
  {
    is_unicode = 0;
    read_direntry(pvd->root_direntry);
  }
  
  if((size + offset) == 0)  // enter root
    read_directory(DVDToc->file[0].sector, DVDToc->file[0].size);
  else
    read_directory(offset>>11, size);

  free(bufferDVD);
  if(files>0)
    return files;
  return NO_FILES;
}

#define IOCTL_DI_READID		0x70
#define IOCTL_DI_READ		0x71
#define IOCTL_DI_WAITCVRCLOSE	0x79
#define IOCTL_DI_GETCOVER	0x88
#define IOCTL_DI_RESET		0x8A
#define IOCTL_DI_OPENPART	0x8B
#define IOCTL_DI_CLOSEPART	0x8C
#define IOCTL_DI_UNENCREAD	0x8D
#define PTABLE_OFFSET	0x40000

static u32 inbuf[ 8 ]  ATTRIBUTE_ALIGN( 32 );
static u32 outbuf[ 8 ] ATTRIBUTE_ALIGN( 32 );

static const char di_fs[] ATTRIBUTE_ALIGN( 32 ) = "/dev/di";
static s32 di_fd = -1;

s32 WDVD_Init(void) {
    if (di_fd < 0) {
	di_fd = IOS_Open(di_fs, 0);
	if (di_fd < 0)
	    return di_fd;
    }

    return 0;
}

s32 WDVD_Close(void) {
    if (di_fd >= 0) {
	IOS_Close(di_fd);
	di_fd = -1;
    }

    return 0;
}

s32 WDVD_Reset(void) {
    s32 ret;

    memset(inbuf, 0, sizeof(inbuf));

    inbuf[0] = IOCTL_DI_RESET << 24;
    inbuf[1] = 1;

    ret = IOS_Ioctl(di_fd, IOCTL_DI_RESET, inbuf, sizeof(inbuf), outbuf, sizeof(outbuf));
    if (ret < 0)
	return ret;

    return (ret == 1) ? 0 : -ret;
}

s32 WDVD_ReadDiskId(void *id) {
    s32 ret;

    memset(inbuf, 0, sizeof(inbuf));

    inbuf[0] = IOCTL_DI_READID << 24;

    ret = IOS_Ioctl(di_fd, IOCTL_DI_READID, inbuf, sizeof(inbuf), outbuf, sizeof(outbuf));
    if (ret < 0)
	return ret;

    if (ret == 1) {
	memcpy(id, outbuf, sizeof(dvddiskid));
	return 0;
    }

    return -ret;
}

s32 WDVD_OpenPartition(u64 offset) {
    u8 *vector = NULL;

    u32 *buffer = NULL;
    s32 ret;

    buffer = (u32 *)memalign(32, 0x5000);
    if (!buffer)
	return -1;

    vector = (u8 *)buffer;

    memset(buffer, 0, 0x5000);

    buffer[0] = (u32)(buffer + 0x10);
    buffer[1] = 0x20;
    buffer[3] = 0x024A;
    buffer[6] = (u32)(buffer + 0x380);
    buffer[7] = 0x49E4;
    buffer[8] = (u32)(buffer + 0x360);
    buffer[9] = 0x20;

    buffer[(0x40 >> 2)]     = IOCTL_DI_OPENPART << 24;
    buffer[(0x40 >> 2) + 1] = offset >> 2;

    ret = IOS_Ioctlv(di_fd, IOCTL_DI_OPENPART, 3, 2, (ioctlv *)vector);

    free(buffer);

    if (ret < 0)
	return ret;

    return (ret == 1) ? 0 : -ret;
}

s32 WDVD_ClosePartition(void) {
    s32 ret;

    memset(inbuf, 0, sizeof(inbuf));

    inbuf[0] = IOCTL_DI_CLOSEPART << 24;

    ret = IOS_Ioctl(di_fd, IOCTL_DI_CLOSEPART, inbuf, sizeof(inbuf), NULL, 0);
    if (ret < 0)
	return ret;

    return (ret == 1) ? 0 : -ret;
}

s32 WDVD_UnencryptedRead(void *buf, u32 len, u64 offset) {
    s32 ret;

    memset(inbuf, 0, sizeof(inbuf));

    inbuf[0] = IOCTL_DI_UNENCREAD << 24;
    inbuf[1] = len;
    inbuf[2] = (u32)(offset >> 2);

    ret = IOS_Ioctl(di_fd, IOCTL_DI_UNENCREAD, inbuf, sizeof(inbuf), buf, len);
    if (ret < 0)
	return ret;

    return (ret == 1) ? 0 : -ret;
}

s32 WDVD_Read(void *buf, u32 len, u64 offset) {
    s32 ret;

    memset(inbuf, 0, sizeof(inbuf));

    inbuf[0] = IOCTL_DI_READ << 24;
    inbuf[1] = len;
    inbuf[2] = (u32)(offset >> 2);

    ret = IOS_Ioctl(di_fd, IOCTL_DI_READ, inbuf, sizeof(inbuf), buf, len);
    if (ret < 0)
	return ret;

    return (ret == 1) ? 0 : -ret;
}

s32 WDVD_WaitForDisc(void) {
    s32 ret;

    memset(inbuf, 0, sizeof(inbuf));

    inbuf[0] = IOCTL_DI_WAITCVRCLOSE << 24;

    ret = IOS_Ioctl(di_fd, IOCTL_DI_WAITCVRCLOSE, inbuf, sizeof(inbuf), outbuf, sizeof(outbuf));
    if (ret < 0)
	return ret;

    return (ret == 1) ? 0 : -ret;
}

s32 WDVD_GetCoverStatus(u32 *status) {
    s32 ret;

    memset(inbuf, 0, sizeof(inbuf));

    inbuf[0] = IOCTL_DI_GETCOVER << 24;

    ret = IOS_Ioctl(di_fd, IOCTL_DI_GETCOVER, inbuf, sizeof(inbuf), outbuf, sizeof(outbuf));
    if (ret < 0)
	return ret;

    if (ret == 1) {
	memcpy(status, outbuf, sizeof(u32));

	return 0;
    }

    return -ret;
}

s32 WDVD_FindPartition(u64 *outbuf)
{
    u64 offset = 0, table_offset = 0;

    u32 cnt, nb_partitions;
    s32 ret;
    u32 *buffer = NULL;

    buffer = (u32 *)memalign(32, 0x20);
    if (!buffer)
	return -1;

    ret = WDVD_UnencryptedRead(buffer, 0x20, PTABLE_OFFSET);
    if (ret < 0)
    {
	free( buffer );
	return ret;
    }

    nb_partitions = buffer[0];
    table_offset  = buffer[1] << 2;

    ret = WDVD_UnencryptedRead(buffer, 0x20, table_offset);
    if (ret < 0)
    {
	free( buffer );
	return ret;
    }

    for (cnt = 0; cnt < nb_partitions; cnt++) {
	u32 type = buffer[cnt * 2 + 1];

	if (!type)
	{
	    offset = buffer[cnt * 2] << 2;
	    break;
	}
    }

    free( buffer );

    if (!offset)
	return -1;

    *outbuf = offset;

    return 0;
}

s32 WDVD_InitAndOpenDataPartition()
{
    u64 dataPartitionOffset;
    u32 status;

    s32 ret = WDVD_Init();
    if( ret < 0 )
    {
	gprintf("WDVD_Init(): %d\n", ret );
	return ret;
    }

    ret = WDVD_GetCoverStatus( &status );
    if( ret < 0 )
    {
	gprintf("WDVD_GetCoverStatus(): %d\n", ret );
	return ret;
    }
    if( status != 2 )
    {
	gprintf("discstatus: %d\n", status );
	return -1;
    }

    ret = WDVD_Reset();
    if( ret < 0 )
    {
	gprintf("WDVD_Reset(): %d\n", ret );
	return ret;
    }

    ret = WDVD_ReadDiskId( (void*)0x80000000 );
    if( ret < 0 )
    {
	gprintf("WDVD_ReadDiskId(): %d\n", ret );
	return ret;
    }

    ret = WDVD_FindPartition( &dataPartitionOffset );
    if( ret < 0 )
    {
	gprintf("WDVD_FindPartition(): %d\n", ret );
	return ret;
    }

    ret = WDVD_OpenPartition( dataPartitionOffset );
    if( ret < 0 )
    {
	gprintf("WDVD_OpenPartition( %016llx ): %d\n", dataPartitionOffset, ret );
	return ret;
    }

    //gprintf("opened partition @ %016llx\n", dataPartitionOffset );

    return 0;
}

s32 WDVD_GetMainDol( void *outbuf, u32* size )
{
    s32 ret;
    FST_INFO fst_info;
    //get dol offset
    ret = WDVD_Read( (u8*)&fst_info, sizeof( FST_INFO ), 0x420LL );
    if( ret < 0 )
    {
	gprintf("WDVD_Read( fst_info ): %d\n", ret );
	return -1;
    }
    fst_info.dol_offset <<= 2;

    //gprintf("dol offset: %08x\n", fst_info.dol_offset );

    //determine dol size
    ret = WDVD_Read( (u8*)outbuf, 0x100, fst_info.dol_offset );
    if( ret < 0 )
    {
	gprintf("WDVD_Read( dol header ): %d\n", ret );
	return -2;
    }
    u32 max = 0;
    u32 i;
    for (i = 0; i < 18; i++) {
	u32 offset = *(u32 *)( outbuf + (i * 4));
	u32 size = *(u32 *)( outbuf + (i * 4) + 0x90);
	if ((offset + size) > max) max = offset + size;
    }
    //gprintf("dol size: %08x\n", max );

    //read the dol
    ret = WDVD_Read( (u8*)outbuf, max, fst_info.dol_offset );
    if( ret < 0 )
    {
	gprintf("WDVD_Read( main.dol ): %d\n", ret );
	return -3;
    }

    *size = max;
    //gprintf("dol read ok\n", max );

    return 0;
}

#include <stdarg.h>
void gprintf(const char *str, ...)
{
	char astr[4096];

	va_list ap;
	va_start(ap,str);

	vsprintf( astr, str, ap );

	va_end(ap);

	usb_flush(1);
	usb_sendbuffer( 1, astr, strlen(astr) );
}

char ascii(char s)
{
    if(s < 0x20)
	return '.';
    if(s > 0x7E)
	return '.';
    return s;
}

void hexdump(void *d, int len)
{
    u8 *data;
    int i, off;
    data = (u8*)d;

    gprintf("\n       0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F  0123456789ABCDEF");
    gprintf("\n====  ===============================================  ================\n");

    for (off=0; off<len; off += 16)
    {
	gprintf("%04x  ",off);
	for(i=0; i<16; i++)
	    if((i+off)>=len)
		gprintf("   ");
	else
	    gprintf("%02x ",data[off+i]);

	gprintf(" ");
	for(i=0; i<16; i++)
	    if((i+off)>=len)
		gprintf(" ");
	    else
		gprintf("%c",ascii(data[off+i]));
	gprintf("\n");
    }
}

s32 LoadFST()
{
//    gprintf("LoadFST(): ");


    s32 ret;
    FST_INFO fst_info;

    ForgetFST();
    //find FST inside partition
    ret = WDVD_Read( (u8*)&fst_info, sizeof( FST_INFO ), 0x420LL );
    if( ret < 0 )
    {
	gprintf("WDVD_Read( fst_info ): %d\n", ret );
	return -1;
    }
    fst_info.fst_offset <<= 2;
    fst_info.fst_size <<= 2;
    //gprintf("fst offset: %08x\tsize: %08x\n", fst_info.fst_offset, fst_info.fst_size );
    //hexdump( &fst_info, 0x20 );

    fstbuffer = (u8*)memalign( 32, fst_info.fst_size );
    if (!fstbuffer)
	return -2;

    //read fst into memory
    ret = WDVD_Read( fstbuffer, fst_info.fst_size, fst_info.fst_offset );
    if( ret < 0 )
    {
	gprintf("WDVD_Read( fst_buffer ): %d\n", ret );
	free( fstbuffer );
	return -3;
    }

    fstLoaded = true;

    //gprintf("OK!\n");
    return 0;
}

void ForgetFST()
{
    if( fstbuffer )
    {
	free( fstbuffer );
	fstbuffer = NULL;
    }
    fstLoaded = false;
}

bool IsFSTLoaded()
{
    return fstLoaded;
}


#include <malloc.h>
#include <stdio.h>
#include <stdlib.h>
#include <ogc/machine/processor.h>
#include <gccore.h>
#include <dirent.h>
#include <string.h>


typedef struct _dolheader {
    u32 text_pos[7];
    u32 data_pos[11];
    u32 text_start[7];
    u32 data_start[11];
    u32 text_size[7];
    u32 data_size[11];
    u32 bss_start;
    u32 bss_size;
    u32 entry_point;
} dolheader;

u32 load_dol(const void *dolstart, struct __argv *argv) {
    u32 i;
    dolheader *dolfile;

    if (dolstart) {
	dolfile = (dolheader *) dolstart;
	for (i = 0; i < 7; i++) {
	    if ((!dolfile->text_size[i]) || (dolfile->text_start[i] < 0x100)) continue;
	    ICInvalidateRange ((void *) dolfile->text_start[i],dolfile->text_size[i]);
	    memcpy((void *) dolfile->text_start[i],dolstart+dolfile->text_pos[i],dolfile->text_size[i]);
	}

	for (i = 0; i < 11; i++) {
	    if ((!dolfile->data_size[i]) || (dolfile->data_start[i] < 0x100)) continue;
	    memcpy((void *) dolfile->data_start[i],dolstart+dolfile->data_pos[i],dolfile->data_size[i]);
	    DCFlushRangeNoSync ((void *) dolfile->data_start[i],dolfile->data_size[i]);
	}

	memset ((void *) dolfile->bss_start, 0, dolfile->bss_size);
	DCFlushRange((void *) dolfile->bss_start, dolfile->bss_size);

	if (argv && argv->argvMagic == ARGV_MAGIC) {
	    void *new_argv = (void *)(dolfile->entry_point + 8);
	    memcpy(new_argv, argv, sizeof(*argv));
	    DCFlushRange(new_argv, sizeof(*argv));
	}

	return dolfile->entry_point;
    }
    return 0;
}


