/*
 * file_dvd.c
 * 
 *  ISO9660/Joliet DVD loading support
 *
 *  Softdev (2006)
 *  Eke-Eke (2007,2008,2009)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 ********************************************************************************/

#if 0

#include "shared.h"
#include "gui.h"
#include "dvd.h"
#include "unzip.h"
#include "filesel.h"
#include "file_fat.h"
#include "file_dvd.h"

/** Minimal ISO Directory Definition **/
#define RECLEN 0            /* Record length */
#define EXTENT 6            /* Extent */
#define FILE_LENGTH 14      /* File length (BIG ENDIAN) */
#define FILE_FLAGS 25       /* File flags */
#define FILENAME_LENGTH 32  /* Filename length */
#define FILENAME 33         /* ASCIIZ filename */

/** Minimal Primary Volume Descriptor **/
#define PVDROOT 0x9c

/** Static Variables **/
static u64 rootdir        = 0;
static u64 basedir        = 0;
static int rootdirlength  = 0;
static int IsJoliet       = 0;
static int diroffset      = 0;
static int haveDVDdir     = 0;
static char dvdbuffer[DVDCHUNK];

/****************************************************************************
 * Primary Volume Descriptor
 *
 * The PVD should reside between sector 16 and 31.
 * This is for single session DVD only.
 ****************************************************************************/
static int getpvd()
{
  int sector = 16;
  u32 rootdir32;

  basedir = rootdirlength = 0;
  IsJoliet = -1;

  /** Look for Joliet PVD first **/
  while (sector < 32)
  {
    if (dvd_read (&dvdbuffer, DVDCHUNK, (u64)(sector << 11)))
    {
      if (memcmp (&dvdbuffer, "\2CD001\1", 8) == 0)
      {
        memcpy(&rootdir32, &dvdbuffer[PVDROOT + EXTENT], 4);
        basedir = (u64)rootdir32;
        memcpy (&rootdirlength, &dvdbuffer[PVDROOT + FILE_LENGTH], 4);
        basedir <<= 11;
        IsJoliet = 1;
        break;
      }
    }
    else
      return 0; /*** Can't read sector! ***/

    sector++;
  }

  if (IsJoliet > 0)
    return 1; /*** Joliet PVD Found ? ***/

  /*** Look for standard ISO9660 PVD ***/
  sector = 16;
  while (sector < 32)
  {
    if (dvd_read (&dvdbuffer, DVDCHUNK, (u64)(sector << 11)))
    {
      if (memcmp (&dvdbuffer, "\1CD001\1", 8) == 0)
      {
        memcpy (&rootdir32, &dvdbuffer[PVDROOT + EXTENT], 4);
        basedir = (u64)rootdir32;
        memcpy (&rootdirlength, &dvdbuffer[PVDROOT + FILE_LENGTH], 4);
        IsJoliet = 0;
        basedir <<= 11;
        break;
      }
    }
    else
      return 0; /*** Can't read sector! ***/
    
    sector++;
  }

  return (IsJoliet == 0);
}

/****************************************************************************
 * getentry
 *
 * Support function to return the next file entry, if any
 * Declared static to avoid accidental external entry.
 ****************************************************************************/
static int getentry(int entrycount)
{
  char fname[512]; /* Huge, but experience has determined this */
  char *ptr;
  char *filename;
  char *filenamelength;
  char *rr;
  int j;
  u32 offset32;

  /* Basic checks */
  if (entrycount >= MAXFILES)
    return 0;
  if (diroffset >= DVDCHUNK)
    return 0;

  /** Decode this entry **/
  if (dvdbuffer[diroffset])  /* Record length available */
  {
    /* Update offsets into sector buffer */
    ptr = (char *) &dvdbuffer[0];
    ptr += diroffset;
    filename = ptr + FILENAME;
    filenamelength = ptr + FILENAME_LENGTH;

    /* Check for wrap round - illegal in ISO spec,
     * but certain crap writers do it! */
    if ((diroffset + dvdbuffer[diroffset]) > DVDCHUNK)
      return 0;

    if (*filenamelength)
    {
      memset (&fname, 0, 512);

      /** Do ISO 9660 first **/
      if (!IsJoliet)
        strcpy (fname, filename);

      else
      {
        /** The more tortuous unicode joliet entries **/
        for (j = 0; j < (*filenamelength >> 1); j++)
          fname[j] = filename[j * 2 + 1];
        fname[j] = 0;

        if (strlen (fname) >= MAXJOLIET)
          fname[MAXJOLIET - 1] = 0;
        if (strlen (fname) == 0)
          fname[0] = filename[0];
      }

      if (strlen (fname) == 0)
        strcpy (fname, ".");
      else
      {
        if (fname[0] == 1)
          strcpy (fname, "..");
        else
        {
          /*
           * Move *filenamelength to t,
           * Only to stop gcc warning for noobs :)
           */
          int t = *filenamelength;
          fname[t] = 0;
        }
      }

      /* Rockridge Check */
      rr = strstr (fname, ";");
      if (rr != NULL)
        *rr = 0;

      strcpy (filelist[entrycount].filename, fname);
      memcpy (&offset32, &dvdbuffer[diroffset + EXTENT], 4);
      filelist[entrycount].offset = (u64)offset32;
      memcpy (&filelist[entrycount].length, &dvdbuffer[diroffset + FILE_LENGTH], 4);
      memcpy (&filelist[entrycount].flags, &dvdbuffer[diroffset + FILE_FLAGS], 1);
      filelist[entrycount].offset <<= 11;
      filelist[entrycount].flags = filelist[entrycount].flags & 2;

      /* Prepare for next entry */
      diroffset += dvdbuffer[diroffset];

      return 1;
    }
  }
  return 0;
}

/***************************************************************************
 * DVD_ClearDirectory
 *
 * Clear DVD directory flag
 ***************************************************************************/ 
void DVD_ClearDirectory(void)
{
  haveDVDdir = 0;
}

/***************************************************************************
 * DVD_UpdateDirectory
 *
 * Update DVD current root directory
 ***************************************************************************/ 
int DVD_UpdateDirectory(bool go_up, u64 offset, u32 length)
{
  /* root has no parent directory */
  if ((basedir == rootdir) && (go_up || (offset == basedir)))
    return 0;

  /* simply update current root directory */
  rootdir = offset;
  rootdirlength = length;

  return 1;
}

/****************************************************************************
 * DVD_ParseDirectory
 *
 * This function will parse the directory tree.
 * It relies on rootdir and rootdirlength being pre-populated by a call to
 * getpvd, a previous parse or a menu selection.
 *
 * The return value is number of files collected, or 0 on failure.
 ****************************************************************************/
int DVD_ParseDirectory(void)
{
  int pdlength;
  u64 pdoffset;
  u64 rdoffset;
  int len = 0;
  int filecount = 0;

  pdoffset = rdoffset = rootdir;
  pdlength = rootdirlength;
  filecount = 0;

  /** Clear any existing values ***/
  memset (&filelist, 0, sizeof (FILEENTRIES) * MAXFILES);

  /*** Get as many files as possible ***/
  while (len < pdlength)
  {
    if (dvd_read (&dvdbuffer, DVDCHUNK, pdoffset) == 0)
      return 0;

    diroffset = 0;

    while (getentry (filecount))
    {
      if (strcmp(filelist[filecount].filename,".") && (filecount < MAXFILES))
        filecount++;
    }

    len += DVDCHUNK;
    pdoffset = rdoffset + len;
  }

  /* Sort the file list */
  qsort(filelist, filecount, sizeof(FILEENTRIES), FileSortCallback);

  return filecount;
}

/****************************************************************************
 * DVD_LoadFile
 *
 * This function will load a BIN, SMD or ZIP file from DVD into the ROM buffer.
 * The index values indicates the file position in filentry list 
 * This functions return the actual size of data copied into the buffer
 *
 ****************************************************************************/ 
int DVD_LoadFile(u8 *buffer, u32 selection)
{
  /* file size */
  int length = filelist[selection].length;

  if (length > 0)
  {
    /* Read first data chunk */
    char readbuffer[DVDCHUNK];
    u64 discoffset = filelist[selection].offset;
    dvd_read (&readbuffer, DVDCHUNK, discoffset);

    /* determine file type */
    if (!IsZipFile ((char *) readbuffer))
    {
      if (length > MAXROMSIZE)
      {
        GUI_WaitPrompt("Error","File size not supported !");
        return 0;
      }

      char msg[64];
      sprintf(msg,"Loading %d bytes...", length);
      GUI_MsgBoxOpen("Information",msg,1);

      /* How many 2k blocks to read */
      int blocks = length / DVDCHUNK;
      int readoffset = 0;
      int i;

      /* read data chunks */
      for (i = 0; i < blocks; i++)
      {
        dvd_read(readbuffer, DVDCHUNK, discoffset);
        discoffset += DVDCHUNK;
        memcpy (buffer + readoffset, readbuffer, DVDCHUNK);
        readoffset += DVDCHUNK;
      }

      /* final read */ 
      i = length % DVDCHUNK;
      if (i)
      {
        dvd_read (readbuffer, DVDCHUNK, discoffset);
        memcpy (buffer + readoffset, readbuffer, i);
      }

      return length;
    }
    else
    {
      return UnZipBuffer (buffer, discoffset, NULL);
    }
  }

  return 0;
}

/****************************************************************************
 * DVD_Open
 *
 * Function to load a DVD directory and display to user.
 ****************************************************************************/ 

int DVD_Open(void)
{
  /* is DVD mounted ? */
  if (!getpvd())
  {
    /* remount DVD */
    GUI_MsgBoxOpen("Information", "Mounting DVD ...",1);
    haveDVDdir = 0;

#ifdef HW_RVL
    u32 val;
    DI_GetCoverRegister(&val);  

    if(val & 0x1)
    {
      GUI_WaitPrompt("Error","No Disc inserted !");
      return 0;
    }

    DI_Mount();
    while(DI_GetStatus() & DVD_INIT) usleep(10);
    if (!(DI_GetStatus() & DVD_READY))
    {
      char msg[64];
      sprintf(msg, "DI Status Error: 0x%08X !\n",DI_GetStatus());
      GUI_WaitPrompt("Error",msg);
      return 0;
    }
#else
    DVD_Mount();
#endif

    if (!getpvd())
    {
      GUI_WaitPrompt("Error","Disc can not be read !");
      return 0;
    }

    /* reset current directory */
    rootdir = basedir;

    GUI_MsgBoxClose();
  }

  if (!haveDVDdir)
  {
    /* parse current directory */
    int max = DVD_ParseDirectory ();

    if (max)
    {
      /* set DVD access flag */
      haveDVDdir = 1;

      /* reset File selector */
      ClearSelector(max);

      /* clear FAT access flag */
      FAT_ClearDirectory();

      return 1;
    }
    else
    {
      /* no entries found */
      GUI_WaitPrompt("Error","No files found !");

      return 0;
    }
  }

  return 1;
}
#endif

#include <errno.h>
#include <ogc/es.h>
#include <ogc/lwp_watchdog.h>
#include <ogcsys.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <sys/dir.h>
#include <sys/iosupport.h>
#include <stdarg.h>


#include "file_dvd.h"

#define DEVICE_NAME "dvd"
#define FST_MAXPATHLEN 255
#define SECTOR_SIZE 0x800

void gprintf(const char *str, ...)
{
	char astr[4096];

	va_list ap;
	va_start(ap,str);

	vsprintf( astr, str, ap );

	va_end(ap);

	usb_flush(1);
	//usb_sendbuffer_safe( 1, astr, strlen(astr) );
	usb_sendbuffer( 1, astr, strlen(astr) );
}

char ascii(char s)
{
    if(s < 0x20)
	return '.';
    if(s > 0x7E)
	return '.';
    return s;
}

void hexdump(void *d, int len)
{
    u8 *data;
    int i, off;
    data = (u8*)d;

    gprintf("\n       0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F  0123456789ABCDEF");
    gprintf("\n====  ===============================================  ================\n");

    for (off=0; off<len; off += 16)
    {
	gprintf("%04x  ",off);
	for(i=0; i<16; i++)
	    if((i+off)>=len)
		gprintf("   ");
	else
	    gprintf("%02x ",data[off+i]);

	gprintf(" ");
	for(i=0; i<16; i++)
	    if((i+off)>=len)
		gprintf(" ");
	    else
		gprintf("%c",ascii(data[off+i]));
	gprintf("\n");
    }
}

typedef struct {
    u32 dol_offset;
    u32 fst_offset;
    u32 fst_size;
    u32 fst_size2;
} __attribute__ ((packed)) FST_INFO;

typedef struct {
    u8 filetype;
    char name_offset[3];
    u32 fileoffset;
    u32 filelen;
} __attribute__((packed)) FST_ENTRY;

typedef struct
{
    bool inUse;
    u32 offset;
    s32 entry;
} OpenFstItem;

//these are the only memory consumed by this devotab
OpenFstItem openDir;
OpenFstItem openFile;
static u8 *fst_buffer = NULL;

//pointer to the name list at the end of fst_buffer
static char *name_table = NULL;
static FST_ENTRY *fst = NULL;

static u64 last_access = 0;
static s32 dotab_device = -1;

const char *FstName( const FST_ENTRY *entry )
{
    if( entry == &fst[ 0 ] )
	return NULL;

    return (char*)name_table + ( *((u32 *)entry) & 0x00ffffff );
}

static int strcmp_slash( const char *s1, const char *s2 )
{
    while( *s1 == *s2++ )
	if( *s1++ == 0 || *s1 == '/' )
	    return( 0 );

    return( *(const unsigned char*)s1 - *(const unsigned char *)( s2 - 1 ) );
}

static int strlen_slash( const char *s )
{
    int ret = 0;
    while( *s && *s++ != '/' )
	ret++;

    return ret;
}

static u32 NextEntryInFolder( u32 current, u32 directory )
{
    u32 next = ( fst[ current ].filetype ? fst[ current ].filelen : current + 1 );
    if( next < fst[ directory ].filelen )
	return next;

    return 0;
}

static s32 EntryFromPath( const char *path, int d )
{

    if( !strncmp( path, "dvd:/", strlen( "dvd:/" ) ) )
	return EntryFromPath( path + 5, 0 );

    if( !fst[ d ].filetype )
    {
	gprintf("ERROR!!  %s is not a directory\n", FstName( &fst[ d ] ) );
	return -1;
    }

    u32 next = d + 1;

    FST_ENTRY *entry = &fst[ next ];

    while( next )
    {
	//gprintf( "\tcomparing: [ %u ] %s\n", next, FstName( entry ) );
	//does this entry match.
	//strlen_slash is used because if looking for "dvd:/gameboy/" it would return a false positive if it hit "dvd:/gameboy advance/" first
	if( !strcmp_slash( path, FstName( entry ) ) && ( strlen( FstName( entry ) ) == strlen_slash( path ) ) )
	{
	    char *slash = strchr( path, '/' );
	    if( slash && strlen( slash + 1 ) )
	    {
		//we are looking for a file somewhere in this folder
		return EntryFromPath( slash + 1, next );
	    }
	    //this is the actual entry we were looking for
	    return next;
	}

	//find the next entry in this folder
	next = NextEntryInFolder( next, d );
	entry = &fst[ next ];
    }

    //no entry with the given path was found
    return -1;
}

static int _FST_open_r( struct _reent *r, void *fileStruct, const char *path, int flags, int mode )
{
    //gprintf("_FST_open_r( \"%s\"\n", path );
    if( openFile.inUse )
    {
	gprintf("tried to open more than 1 file at a time\n");
	r->_errno = ENOENT;
	return -1;
    }
    openFile.entry = EntryFromPath( path, 0 );
    if( openFile.entry < 0 )
    {
	//gprintf("invalid filename: %s\n", path );
	r->_errno = ENOENT;
	return -1;
    }
    if( fst[ openFile.entry ].filetype )
    {
	r->_errno = EISDIR;
	return -1;
    }
    openFile.offset = 0;
    openFile.inUse = true;

    return 1;
}

static int _FST_close_r( struct _reent *r, int fd )
{
    //gprintf("close\n");
    if( !openFile.inUse )
    {
	gprintf("tried to close a file that isnt open\n" );
	r->_errno = EBADF;
	return -1;
    }
    memset( &openFile, 0, sizeof( OpenFstItem ) );
    return 0;
}

static int _FST_read_r( struct _reent *r, int fd, char *ptr, size_t len )
{
    //gprintf("read( %d )\n", fd );
    if( !openFile.inUse )
    {
	gprintf("read: !openFile.inUse\n");
	r->_errno = EBADF;
	return -1;
    }
    if( openFile.offset >= fst[ openFile.entry ].filelen )
    {
	//gprintf("read: 2\n");
	r->_errno = EOVERFLOW;
	return 0;
    }
    if( len + openFile.offset > fst[ openFile.entry ].filelen )
    {
	//gprintf("read: 3\n");
	r->_errno = EOVERFLOW;
	len = fst[ openFile.entry ].filelen - openFile.offset;
    }
    if( len <= 0 )
    {
	return 0;
    }

    if( WDVD_Read_Safe( (u8*)ptr, len, (u64)( ( (u64)fst[ openFile.entry ].fileoffset << 2LL) + openFile.offset ) ) < 0 )
    {
	gprintf("read: failed\n");
	r->_errno = EIO;
	return -1;
    }

    openFile.offset += len;

    return len;

}

static off_t _FST_seek_r( struct _reent *r, int fd, off_t pos, int dir )
{
    //gprintf("_FST_seek_r()\n" );

    if( !openFile.inUse )
    {
	gprintf("seek: !openFile.inUse\n");
	r->_errno = EBADF;
	return -1;
    }

    int position;

    switch( dir ) {
	case SEEK_SET:
	    position = pos;
	    break;
	case SEEK_CUR:
	    position = openFile.offset + pos;
	    break;
	case SEEK_END:
	    position = fst[ openFile.entry ].filelen + pos;
	    break;
	default:
	    r->_errno = EINVAL;
	    return -1;
    }

    if( pos > 0 && position < 0 )
    {
	gprintf("seek: pos > 0 && position < 0\n");
	r->_errno = EOVERFLOW;
	return -1;
    }

    if( position < 0 || position > fst[ openFile.entry ].filelen )
    {
	gprintf("seek: shit\n");
	r->_errno = EINVAL;
	return -1;
    }
    openFile.offset = position;

    return position;
}

static void stat_entry_( u32 entry, struct stat *st) {
    st->st_dev = 0x4657;
    st->st_ino = 0;
    st->st_mode = ( fst[ entry ].filetype ? S_IFDIR : S_IFREG) | (S_IRUSR | S_IRGRP | S_IROTH);
    st->st_nlink = 1;
    st->st_uid = 1;
    st->st_gid = 2;
    st->st_rdev = st->st_dev;
    st->st_size = fst[ entry ].filelen;
    st->st_atime = 0;
    st->st_spare1 = 0;
    st->st_mtime = 0;
    st->st_spare2 = 0;
    st->st_ctime = 0;
    st->st_spare3 = 0;
    st->st_blksize = SECTOR_SIZE;
    st->st_blocks = ( fst[ entry ].filelen + SECTOR_SIZE - 1) / SECTOR_SIZE;
    st->st_spare4[0] = 0;
    st->st_spare4[1] = 0;
}

static int _FST_fstat_r(struct _reent *r, int fd, struct stat *st) {
    gprintf("fstat()\n" );

    if( !openFile.inUse )
    {
	r->_errno = EBADF;
	return -1;
    }
    stat_entry_( openFile.entry, st );

    return 0;
}

static int _FST_stat_r(struct _reent *r, const char *path, struct stat *st) {
    //gprintf("stat( \"%s\" )\n", path );
    s32 er = EntryFromPath( path, 0 );
    if( er < 0 )
    {
	r->_errno = ENOENT;
	return -1;
    }
    stat_entry_( er, st );

    return 0;
}

static int _FST_chdir_r(struct _reent *r, const char *path) {
    //gprintf("chdir( \"%s\" )\n", path );
    s32 er = EntryFromPath( path, 0 );
    if( er < 0 )
    {
	gprintf("chdir() bad directory name   %s\n", path );
	r->_errno = ENOTDIR;
	return -1;
    }
    if( !fst[ er ].filetype )
    {
	gprintf("chdir() !filetype\n");
	r->_errno = ENOTDIR;
	return -1;
    }

    memset( &openDir.inUse, 0, sizeof( OpenFstItem ) );
    openDir.entry = er;
    openDir.offset = openDir.entry ? openDir.entry : openDir.entry + 1;
    openDir.inUse = true;

    return 0;
}

static DIR_ITER *_FST_diropen_r(struct _reent *r, DIR_ITER *dirState, const char *path)
{
    //gprintf("dir open( \"%s\" )\n", path );
    if( openDir.inUse )
    {
	gprintf("tried to open more than 1 directory\n");
	r->_errno = ENOENT;
	return NULL;
    }
    if( !strcmp( path, "dvd:/" ) )
	openDir.entry = 0;
    else
	openDir.entry = EntryFromPath( path, 0 );

    if( openDir.entry < 0 )
    {
	gprintf("bad filename\n");
	r->_errno = ENOENT;
	return NULL;
    }

    if( !fst[ openDir.entry ].filetype )
    {
	gprintf("not a directory: %s\n", path );
	r->_errno = ENOTDIR;
	return NULL;
    }
    //gprintf("returns %d: %s\n", openDir.entry, FstName( &fst[ openDir.entry ] ) );


    openDir.offset = openDir.entry ? openDir.entry : openDir.entry + 1;

    openDir.inUse = true;
    dirState = (DIR_ITER*)&openDir;

    return dirState;
}

static int _FST_dirreset_r( struct _reent *r, DIR_ITER *dirState )
{
    //gprintf("dirreset()\n" );
    if( !openDir.inUse )
    {
	r->_errno = EBADF;
	return -1;
    }
    openDir.offset = openDir.entry ? openDir.entry : openDir.entry + 1;
    return 0;
}

static int _FST_dirnext_r( struct _reent *r, DIR_ITER *dirState, char *filename, struct stat *st )
{
    if( !openDir.inUse )
    {
	gprintf("dirnext(): !inUse\n");
	r->_errno = EBADF;
	return -1;
    }
    if( openDir.entry && openDir.offset == openDir.entry )//this is the first entry in a directory, and not the root directory
    {
	strcpy( filename, ".." );
	stat_entry_( fst[ openDir.offset ].fileoffset, st );
	openDir.offset = fst[ openDir.entry ].filelen > openDir.entry + 1 ? openDir.offset + 1 : 0;
	return 0;
    }
    if( !openDir.offset )//no more entries in this directory
    {
	r->_errno = ENOENT;
	return -1;
    }

    //gprintf("dirnext: %d: %s\n", openDir.offset, FstName( &fst[ openDir.offset ] ) );

    strncpy( filename, FstName( &fst [ openDir.offset ] ), FST_MAXPATHLEN - 1 );
    stat_entry_( openDir.offset, st );
    openDir.offset = NextEntryInFolder( openDir.offset, openDir.entry );

    return 0;
}

static int _FST_dirclose_r( struct _reent *r, DIR_ITER *dirState )
{
    if( !openDir.inUse )
    {
	gprintf("dirclose(): !openDir.inUse\n");
	r->_errno = EBADF;
	return -1;
    }

    memset( &openDir.inUse, 0, sizeof( OpenFstItem ) );

    return 0;
}

static int _FST_statvfs_r(struct _reent *r, const char *name, struct statvfs *buf)
{
    if(!buf)
	return -1;

    memset(buf, 0,  sizeof(struct statvfs));
    return 0;
}

static const devoptab_t dotab_fst = {
    DEVICE_NAME,
    sizeof(OpenFstItem),
    _FST_open_r,
    _FST_close_r,
    NULL,
    _FST_read_r,
    _FST_seek_r,
    _FST_fstat_r,
    _FST_stat_r,
    NULL,
    NULL,
    _FST_chdir_r,
    NULL,
    NULL,
    sizeof(OpenFstItem),
    _FST_diropen_r,
    _FST_dirreset_r,
    _FST_dirnext_r,
    _FST_dirclose_r,
    _FST_statvfs_r
};

static bool read_disc() {
    //gprintf("read_disc()\n");

    s32 ret;
    FST_INFO fst_info;

    //find FST inside partition
    ret = WDVD_Read( (u8*)&fst_info, sizeof( FST_INFO ), 0x420LL );
    if( ret < 0 )
    {
	gprintf("WDVD_Read( fst_info ): %d\n", ret );
	goto error;
    }
    fst_info.fst_offset <<= 2;
    fst_info.fst_size <<= 2;
    //gprintf("fst offset: %08x\tsize: %08x\n", fst_info.fst_offset, fst_info.fst_size );
    //hexdump( &fst_info, 0x20 );

    fst_buffer = (u8*)memalign( 32, fst_info.fst_size );
    if( !fst_buffer ) goto error;

    //read fst into memory
    ret = WDVD_Read( fst_buffer, fst_info.fst_size, fst_info.fst_offset );
    if( ret < 0 )
    {
	gprintf("WDVD_Read( fst_buffer ): %d\n", ret );
	goto error;
    }

    //set the pointers
    fst = (FST_ENTRY *)fst_buffer;
    u32 name_table_offset = fst->filelen * 0xC;
    name_table = (char *)( fst_buffer + name_table_offset );

    //clear the openFile and openDir entries
    memset( &openDir, 0, sizeof( OpenFstItem ) );
    memset( &openFile, 0, sizeof( OpenFstItem ) );

    //gprintf("read_disc ok\n");
    return true;

    error:
    gprintf("read_disc fail\n");
    if( fst_buffer )
	free( fst_buffer );

    fst = NULL;
    name_table = NULL;

    return false;
}

static bool alreadyMounted = false;

bool FSTIsMounted()
{
    return alreadyMounted;
}

bool FST_Mount()
{
    //gprintf("FST_Mount()\n");

    if( alreadyMounted )
	FST_Unmount();

    bool success = read_disc() && ( dotab_device = AddDevice( &dotab_fst ) ) >= 0;
    if( success )
	last_access = gettime();
    else
	FST_Unmount();

    alreadyMounted = true;
    //gprintf("okey dokey\n");
    return success;
}

bool FST_Unmount()
{
    //gprintf("FST_Unmount()\n");

    alreadyMounted = false;
    if( fst_buffer )
    {
	free( fst_buffer );
	fst_buffer = NULL;
    }
    name_table = NULL;
    fst = NULL;

    last_access = 0;
    if (dotab_device >= 0) {
	dotab_device = -1;
	return !RemoveDevice(DEVICE_NAME ":");
    }

    return true;
}

u64 FST_LastAccess() {
    return last_access;
}

#define IOCTL_DI_READID		0x70
#define IOCTL_DI_READ		0x71
#define IOCTL_DI_WAITCVRCLOSE	0x79
#define IOCTL_DI_GETCOVER	0x88
#define IOCTL_DI_RESET		0x8A
#define IOCTL_DI_OPENPART	0x8B
#define IOCTL_DI_CLOSEPART	0x8C
#define IOCTL_DI_UNENCREAD	0x8D
#define PTABLE_OFFSET	0x40000

#define CACHE_END ( cacheStart + CACHE_SIZE )
#define CACHE_SIZE 0x800
static u8 readCache[ CACHE_SIZE ] ATTRIBUTE_ALIGN( 32 );
static u64 cacheStart = 0;

static u32 inbuf[ 8 ]  ATTRIBUTE_ALIGN( 32 );
static u32 outbuf[ 8 ] ATTRIBUTE_ALIGN( 32 );

static const char di_fs[] ATTRIBUTE_ALIGN( 32 ) = "/dev/di";
static s32 di_fd = -1;

s32 WDVD_Init(void) {
    if (di_fd < 0) {
	di_fd = IOS_Open(di_fs, 0);
	if (di_fd < 0)
	    return di_fd;
    }

    return 0;
}

s32 WDVD_Close(void) {
    if (di_fd >= 0) {
	IOS_Close(di_fd);
	di_fd = -1;
    }

    return 0;
}

s32 WDVD_Reset(void) {
    s32 ret;

    memset(inbuf, 0, sizeof(inbuf));

    inbuf[0] = IOCTL_DI_RESET << 24;
    inbuf[1] = 1;

    ret = IOS_Ioctl(di_fd, IOCTL_DI_RESET, inbuf, sizeof(inbuf), outbuf, sizeof(outbuf));
    if (ret < 0)
	return ret;

    return (ret == 1) ? 0 : -ret;
}

s32 WDVD_ReadDiskId(void *id) {
    s32 ret;

    memset(inbuf, 0, sizeof(inbuf));

    inbuf[0] = IOCTL_DI_READID << 24;

    ret = IOS_Ioctl(di_fd, IOCTL_DI_READID, inbuf, sizeof(inbuf), outbuf, sizeof(outbuf));
    if (ret < 0)
	return ret;

    if (ret == 1) {
	memcpy(id, outbuf, sizeof(dvddiskid));
	return 0;
    }

    return -ret;
}

s32 WDVD_OpenPartition(u64 offset) {
    u8 *vector = NULL;

    u32 *buffer = NULL;
    s32 ret;

    buffer = (u32 *)memalign(32, 0x5000);
    if (!buffer)
	return -1;

    vector = (u8 *)buffer;

    memset(buffer, 0, 0x5000);

    buffer[0] = (u32)(buffer + 0x10);
    buffer[1] = 0x20;
    buffer[3] = 0x024A;
    buffer[6] = (u32)(buffer + 0x380);
    buffer[7] = 0x49E4;
    buffer[8] = (u32)(buffer + 0x360);
    buffer[9] = 0x20;

    buffer[(0x40 >> 2)]     = IOCTL_DI_OPENPART << 24;
    buffer[(0x40 >> 2) + 1] = offset >> 2;

    ret = IOS_Ioctlv(di_fd, IOCTL_DI_OPENPART, 3, 2, (ioctlv *)vector);

    free(buffer);

    if (ret < 0)
	return ret;

    return (ret == 1) ? 0 : -ret;
}

s32 WDVD_ClosePartition(void) {
    s32 ret;

    memset(inbuf, 0, sizeof(inbuf));

    inbuf[0] = IOCTL_DI_CLOSEPART << 24;

    ret = IOS_Ioctl(di_fd, IOCTL_DI_CLOSEPART, inbuf, sizeof(inbuf), NULL, 0);
    if (ret < 0)
	return ret;

    return (ret == 1) ? 0 : -ret;
}

s32 WDVD_UnencryptedRead(void *buf, u32 len, u64 offset) {
    s32 ret;

    memset(inbuf, 0, sizeof(inbuf));

    inbuf[0] = IOCTL_DI_UNENCREAD << 24;
    inbuf[1] = len;
    inbuf[2] = (u32)(offset >> 2);

    ret = IOS_Ioctl(di_fd, IOCTL_DI_UNENCREAD, inbuf, sizeof(inbuf), buf, len);
    if (ret < 0)
	return ret;

    return (ret == 1) ? 0 : -ret;
}

s32 WDVD_Read(void *buf, u32 len, u64 offset) {
    s32 ret;

    memset(inbuf, 0, sizeof(inbuf));

    inbuf[0] = IOCTL_DI_READ << 24;
    inbuf[1] = len;
    inbuf[2] = (u32)(offset >> 2);

    ret = IOS_Ioctl(di_fd, IOCTL_DI_READ, inbuf, sizeof(inbuf), buf, len);
    if (ret < 0)
	return ret;

    return (ret == 1) ? 0 : -ret;
}

s32 WDVD_ReadIntoCache( u64 offset )
{
    u8 off = 0;
    while( ( ( ( offset - off ) >> 2ull ) << 2ull ) != offset - off )
    {
	off++;
    }

    s32 ret = WDVD_Read( &readCache, CACHE_SIZE, offset - off );
    if( ret < 0 )
    {
	gprintf("WDVD_Read(): %d\n", ret);
	return ret;
    }
    cacheStart = offset - off;

    return 0;
}

s32 WDVD_Read_Safe( void *buf, u32 len, u64 offset )
{
    u8 off = 0;
    while( ( ( ( offset - off ) >> 2ull ) << 2ull ) != offset - off )
    {
	off++;
    }
    if( !off && len >= CACHE_SIZE ) //nice happy aligned offset and not just 4 bytes at a time, just read it and be done
	return WDVD_Read( buf, len, offset );

    //stupid bullshit "read 1 byte at a time" shit like 7z does
    u32 haveRead = 0;
    u32 howFarIntoTheCacheIsIt = 0;
    u32 howMuchOfTheCachDoINeedToCopy = 0;
    while( haveRead < len )
    {
	if( offset >= cacheStart && offset < cacheStart + CACHE_SIZE )
	{
	    howFarIntoTheCacheIsIt = (u32)( (u64)offset - (u64)cacheStart );
	    howMuchOfTheCachDoINeedToCopy = MIN( CACHE_END - howFarIntoTheCacheIsIt, len );

	    memcpy( buf + haveRead, (u8*)(&readCache) + howFarIntoTheCacheIsIt, howMuchOfTheCachDoINeedToCopy );
	}
	else
	{
	    s32 ret = WDVD_ReadIntoCache( offset + haveRead );
	    if( ret < 0 )
		return ret;
	}
	haveRead += howMuchOfTheCachDoINeedToCopy;
    }
    return 0;
}

s32 WDVD_WaitForDisc(void) {
    s32 ret;

    memset(inbuf, 0, sizeof(inbuf));

    inbuf[0] = IOCTL_DI_WAITCVRCLOSE << 24;

    ret = IOS_Ioctl(di_fd, IOCTL_DI_WAITCVRCLOSE, inbuf, sizeof(inbuf), outbuf, sizeof(outbuf));
    if (ret < 0)
	return ret;

    return (ret == 1) ? 0 : -ret;
}

s32 WDVD_GetCoverStatus(u32 *status) {
    s32 ret;

    memset(inbuf, 0, sizeof(inbuf));

    inbuf[0] = IOCTL_DI_GETCOVER << 24;

    ret = IOS_Ioctl(di_fd, IOCTL_DI_GETCOVER, inbuf, sizeof(inbuf), outbuf, sizeof(outbuf));
    if (ret < 0)
	return ret;

    if (ret == 1) {
	memcpy(status, outbuf, sizeof(u32));

	return 0;
    }

    return -ret;
}

s32 WDVD_FindPartition(u64 *outbuf)
{
    u64 offset = 0, table_offset = 0;

    u32 cnt, nb_partitions;
    s32 ret;
    u32 *buffer = NULL;

    buffer = (u32 *)memalign(32, 0x20);
    if (!buffer)
	return -1;

    ret = WDVD_UnencryptedRead(buffer, 0x20, PTABLE_OFFSET);
    if (ret < 0)
    {
	free( buffer );
	return ret;
    }

    nb_partitions = buffer[0];
    table_offset  = buffer[1] << 2;

    ret = WDVD_UnencryptedRead(buffer, 0x20, table_offset);
    if (ret < 0)
    {
	free( buffer );
	return ret;
    }

    for (cnt = 0; cnt < nb_partitions; cnt++) {
	u32 type = buffer[cnt * 2 + 1];

	if (!type)
	{
	    offset = buffer[cnt * 2] << 2;
	    break;
	}
    }

    free( buffer );

    if (!offset)
	return -1;

    *outbuf = offset;

    return 0;
}

s32 WDVD_InitAndOpenDataPartition()
{
    u64 dataPartitionOffset;
    u32 status = 0;

    s32 ret = WDVD_Init();
    if( ret < 0 )
    {
	gprintf("WDVD_Init(): %d\n", ret );
	return ret;
    }

    ret = WDVD_GetCoverStatus( &status );
    if( ret < 0 )
    {
	gprintf("WDVD_GetCoverStatus(): %d\n", ret );
	return ret;
    }
    if( status != 2 )
    {
	gprintf("discstatus: %d\n", status );
	return -1;
    }

    ret = WDVD_Reset();
    if( ret < 0 )
    {
	gprintf("WDVD_Reset(): %d\n", ret );
	return ret;
    }

    ret = WDVD_ReadDiskId( (void*)0x80000000 );
    if( ret < 0 )
    {
	gprintf("WDVD_ReadDiskId(): %d\n", ret );
	return ret;
    }

    ret = WDVD_FindPartition( &dataPartitionOffset );
    if( ret < 0 )
    {
	gprintf("WDVD_FindPartition(): %d\n", ret );
	return ret;
    }

    ret = WDVD_OpenPartition( dataPartitionOffset );
    if( ret < 0 )
    {
	gprintf("WDVD_OpenPartition( %016llx ): %d\n", dataPartitionOffset, ret );
	return ret;
    }

    //gprintf("opened partition @ %016llx\n", dataPartitionOffset );

    return 0;
}

s32 WDVD_GetMainDol( void *outbuf, u32* size )
{
    s32 ret;
    FST_INFO fst_info;
    //get dol offset
    ret = WDVD_Read( (u8*)&fst_info, sizeof( FST_INFO ), 0x420LL );
    if( ret < 0 )
    {
	gprintf("WDVD_Read( fst_info ): %d\n", ret );
	return -1;
    }
    fst_info.dol_offset <<= 2;

    //gprintf("dol offset: %08x\n", fst_info.dol_offset );

    //determine dol size
    ret = WDVD_Read( (u8*)outbuf, 0x100, fst_info.dol_offset );
    if( ret < 0 )
    {
	gprintf("WDVD_Read( dol header ): %d\n", ret );
	return -2;
    }
    u32 max = 0;
    u32 i;
    for (i = 0; i < 18; i++) {
	u32 offset = *(u32 *)( outbuf + (i * 4));
	u32 size = *(u32 *)( outbuf + (i * 4) + 0x90);
	if ((offset + size) > max) max = offset + size;
    }
    //gprintf("dol size: %08x\n", max );

    //read the dol
    ret = WDVD_Read( (u8*)outbuf, max, fst_info.dol_offset );
    if( ret < 0 )
    {
	printf("WDVD_Read( main.dol ): %d\n", ret );
	return -3;
    }

    *size = max;
    //gprintf("dol read ok\n", max );

    return 0;
}
