#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <gctypes.h>
#include <ogc/system.h>
#include <fat.h>
#include <wiiuse/wpad.h>
#include <malloc.h>
#include <sys/iosupport.h>
#include <ogc/machine/processor.h>
#include <vector>
#include <string>


u8 *homebrewbuffer = (u8 *)0x92000000;
u32 homebrewsize = 0;
extern const u8		app_booter_dol[];
extern const u32	app_booter_dol_size;
extern "C" {
extern void __exception_closeall();
typedef void (*entrypoint) (void);
}


typedef struct _dolheader {
    u32 text_pos[7];
    u32 data_pos[11];
    u32 text_start[7];
    u32 data_start[11];
    u32 text_size[7];
    u32 data_size[11];
    u32 bss_start;
    u32 bss_size;
    u32 entry_point;
} dolheader;

u32 load_dol(const void *dolstart, struct __argv *argv) {
    u32 i;
    dolheader *dolfile;

    if (dolstart) {
	dolfile = (dolheader *) dolstart;
	for (i = 0; i < 7; i++) {
	    if ((!dolfile->text_size[i]) || (dolfile->text_start[i] < 0x100)) continue;
	    ICInvalidateRange ((void *) dolfile->text_start[i],dolfile->text_size[i]);
	    memcpy((void *) dolfile->text_start[i],dolstart+dolfile->text_pos[i],dolfile->text_size[i]);
	}

	for (i = 0; i < 11; i++) {
	    if ((!dolfile->data_size[i]) || (dolfile->data_start[i] < 0x100)) continue;
	    memcpy((void *) dolfile->data_start[i],dolstart+dolfile->data_pos[i],dolfile->data_size[i]);
	    DCFlushRangeNoSync ((void *) dolfile->data_start[i],dolfile->data_size[i]);
	}

	memset ((void *) dolfile->bss_start, 0, dolfile->bss_size);
	DCFlushRange((void *) dolfile->bss_start, dolfile->bss_size);

	if (argv && argv->argvMagic == ARGV_MAGIC) {
	    void *new_argv = (void *)(dolfile->entry_point + 8);
	    memcpy(new_argv, argv, sizeof(*argv));
	    DCFlushRange(new_argv, sizeof(*argv));
	}

	return dolfile->entry_point;
    }
    return 0;
}

static std::vector<std::string> Arguments;

extern "C" void AddBootArgument(const char * argv)
{
    std::string arg(argv);
    Arguments.push_back(arg);
}

static int SetupARGV(struct __argv * args)
{
    if(!args)
	return -1;

    bzero(args, sizeof(struct __argv));
    args->argvMagic = ARGV_MAGIC;

    u32 stringlength = 1;

    for(u32 i = 0; i < Arguments.size(); i++)
    {
	stringlength += Arguments[i].size()+1;
    }

    args->length = stringlength;
    args->commandLine = (char*) malloc(args->length);
    if (!args->commandLine)
	return -1;

    u32 argc = 0;
    u32 position = 0;

    for(u32 i = 0; i < Arguments.size(); i++)
    {
	strcpy(&args->commandLine[position], Arguments[i].c_str());
	position += Arguments[i].size()+1;
	argc++;
    }

    args->argc = argc;

    args->commandLine[args->length - 1] = '\0';
    args->argv = &args->commandLine;
    args->endARGV = args->argv + 1;

    Arguments.clear();

    return 0;
}

extern "C" int BootHomebrew()
{
    if( homebrewsize == 0 )
	return -1;

    struct __argv args;
    SetupARGV(&args);

    u32 cpu_isr;

    entrypoint entry = (entrypoint)load_dol( app_booter_dol, &args );

    if ( !entry )
	return -1;

    SYS_ResetSystem( SYS_SHUTDOWN, 0, 0 );
    _CPU_ISR_Disable( cpu_isr );
    __exception_closeall();
    entry();
    _CPU_ISR_Restore ( cpu_isr );

    return 0;
}
