/****************************************************************************
 *  ggentry.c
 *
 *  Genesis Plus GX Game Genie 
 *
 *  Softdev (2006)
 *  Eke-Eke (2007,2008,2009)
 *  giantpune (2010)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 ***************************************************************************/

#include "shared.h"
#include "font.h"
#include "gui.h"
#include "menu.h"

#ifdef HW_RVL
#include <wiiuse/wpad.h>
#endif

#define MAX_GGP_ENTRIES 325 //294 is the most i saw in the book + comments ( mortal kombat )
#define PAGESIZE 11
#define PAGEOFFSET 120

static int string_offset = 0;
static void selector_cb(void);

typedef struct
{
  int address;
  unsigned short data;
  unsigned short data_original;//remember the previous value so the code can be disabled
} GGPATCH;

typedef struct
{
  u8 type; //0 = comment, 1 = code
  char text[ 256 ];
  u8 numCodes;
  GGPATCH code[ 5 ]; //support 5 codes per line ( thats the most i got from the official game genie book )
  u8 isEnabled;
} GGPENTRY;

GGPENTRY ggpEntry[ MAX_GGP_ENTRIES ];
static u32 num_ggp	= 0;
static int offset	= 0;
static int selection	= 0;
char ggvalidchars[]	= "ABCDEFGHJKLMNPRSTVWXYZ0123456789*";

void CheatSelector();


static u8 genieToAddrOffset(char *code, int *addr, unsigned short * val)
{
  char *p;
  int n, i;
  int retAddr = 0;
  unsigned short retVal = 0;
  for (i = 0; i < 8; i++)
  {
    /*** This should only happen if memory is corrupt! ***/
    p = strchr (ggvalidchars, code[i]);
    if( p == NULL )
    {
	*addr = 0;
	*val = 0;
	return 0;
    }

    n = p - ggvalidchars;

    switch (i)
    {
      case 0:
      retVal |= n << 3;
      break;

      case 1:
      retVal |= n >> 2;
      retAddr |= (n & 3) << 14;
      break;

      case 2:
      retAddr |= n << 9;
      break;

      case 3:
      retAddr |= (n & 0xF) << 20 | (n >> 4) << 8;
      break;

      case 4:
      retVal |= (n & 1) << 12;
      retAddr |= (n >> 1) << 16;
      break;

      case 5:
      retVal |= (n & 1) << 15 | (n >> 1) << 8;
      break;

      case 6:
      retVal |= (n >> 3) << 13;
      retAddr |= (n & 7) << 5;
      break;

      case 7:
      retAddr |= n;
      break;
    }
  }
  *addr = retAddr;
  *val = retVal;
  return 1;
}

/****************************************************************************
 * ClearGGCodes
 *
 * Should be called whenever a new rom is loaded
 ****************************************************************************/
void ClearGGCodes ()
{
    //gprintf("Clearing cheats\n");
    selection = 0;
    offset = 0;
    num_ggp = 0;
    memset( &ggpEntry, 0, sizeof( GGPENTRY ) * MAX_GGP_ENTRIES );
}

/****************************************************************************
 * GetGGEntries
 *
 * Screen to return encoded Game Genie codes.
 * No keyboard is available, so it's just a simple wrap round each line kind
 * of thing. 
 ****************************************************************************/

static s32 ReadGGPfile( const char* path )
{
    //gprintf("ReadGGPfile( \"%s\" )\n", path );

    u8 *buffer = NULL;
    u32 filesize = 0;

    FILE* file = fopen( path, "rb" );
    if( !file )
	return -1;

    fseek ( file , 0, SEEK_END);
    filesize = ftell( file );
    rewind( file );
    buffer = (u8*)malloc( filesize );
    if ( !buffer )
    {
	fclose( file );
	return -2;
    }

    memset( buffer, 0, filesize );
    if ( fread ( buffer, 1, filesize, file ) != filesize )
    {
	fclose( file );
	free( buffer );
	return -3;
    }
    fclose( file );

    //clear all ggp entries
    memset( &ggpEntry, 0, sizeof( GGPENTRY ) * MAX_GGP_ENTRIES );
    num_ggp = 0;

    char * linebreak = strtok((char *)buffer, "\n");

    while(linebreak != NULL)
    {
	char *tbuf=linebreak;
	//gprintf("cheatline: \"%s\"\n", tbuf );

	if( tbuf[ 0 ] == 'N' )//note
	{
	    tbuf += 2;
	    snprintf( ggpEntry[ num_ggp ].text, sizeof( ggpEntry[ num_ggp ].text ), "%s", tbuf );
	    num_ggp++;
	}
	else if( tbuf[ 0 ] == 'G' )//game-genie code
	{
	    tbuf += 2;

	    u8 fail = 0;
	    GGPENTRY tmp_ggp;
	    memset( &tmp_ggp, 0, sizeof( GGPENTRY ) );

	    tmp_ggp.type = 1;

	    while( 1 )
	    {
		if( tbuf[ 0 ] == ':' )//end of the codes, this is the start of the description
		{
		    tbuf++;
		    break;
		}
		if( tbuf[ 0 ] == '+' )//another code came before this one
		{
		    tbuf++;
		}
		if( tbuf[ 4 ] != '-')//invalid code
		{
		    fail = 1;
		    break;
		}
		if( tmp_ggp.numCodes > 5 )//too many lines of code
		{
		    fail = 1;
		    break;
		}

		char thiscode[10];
		int tmpAddr;
		unsigned short tmpVal;
		memset (&thiscode, 0, 10);

		memcpy ( &thiscode, tbuf, 4 );
		memcpy ( &thiscode[ 4 ], tbuf + 5, 4 );

		u8 r = genieToAddrOffset( (char*)&thiscode, &tmpAddr, &tmpVal );
		//decodeGenesis( (char*)&thiscode, NULL, NULL );
		if( !r )
		{
		    gprintf("failed to convert \"%s\"\n", thiscode );
		    fail = 1;
		    break;
		}
		//gprintf("%s: %08x, %04x\n", thiscode, tmpAddr, tmpVal );
		if( tmpAddr > 0x400000 )
		{
		    gprintf("Address is too high! %08x\n", tmpAddr );
		    fail = 1;
		    break;
		}

		tmp_ggp.code[ tmp_ggp.numCodes ].address = tmpAddr;
		tmp_ggp.code[ tmp_ggp.numCodes ].data = tmpVal;
		if( cart.rom )
		    tmp_ggp.code[ tmp_ggp.numCodes ].data_original = *(uint16 *)(cart.rom + tmpAddr) & 0xffff;

		tmp_ggp.numCodes++;


		tbuf += 9;
	    }
	    if( fail )
	    {
		gprintf("failed!\n");
	    }
	    else
	    {
		snprintf( tmp_ggp.text, sizeof( tmp_ggp.text ), "%s", tbuf );
		memcpy( &ggpEntry[ num_ggp ], &tmp_ggp, sizeof( GGPENTRY ) );
		num_ggp++;
	    }
	}
	//else hexdump( tbuf, strlen( tbuf ) );

	// find next line break
	linebreak = strtok( NULL, "\n" );
    }
    free( buffer );

    //just to help debugging
    /*int k;
    for( k = 0; k < num_ggp; k++ )
    {
	if( !ggpEntry[ k ].type )
	{
	    gprintf("NOTE::\"%s\"\n", ggpEntry[ k ].text );
	}
	else
	{
	    gprintf("CODE::\"%s\"\n", ggpEntry[ k ].text );
	    int p;
	    for( p = 0; p < ggpEntry[ k ].numCodes; p++ )
	    {
		gprintf("\t%08x: %04x -> %04x\n", ggpEntry[ k ].code[ p ].address, ggpEntry[ k ].code[ p ].data_original, ggpEntry[ k ].code[ p ].data );
	    }
	}
    }*/

    return 0;
}
static void patch16( int addr, u16 val )
{
//    gprintf("patch16( %08x, %04x )\n", addr, val );

    *(uint16 *)( cart.rom + addr ) = val & 0xffff;
    DCFlushRange( (void*)( cart.rom + addr ), 4 );
}

static void ApplyCheats()
{
    if( !cart.rom )
    {
	gprintf("tried to apply cheats with no rom loaded\n");
	return;
    }
    int i;

    //first remove all cheats
    for( i = 0; i < num_ggp; i++ )
    {
	if( !ggpEntry[ i ].type )//skip comments
	    continue;

	u8 j;
	for( j = 0; j < ggpEntry[ i ].numCodes; j++ )
	{
	    patch16( ggpEntry[ i ].code[ j ].address, ggpEntry[ i ].code[ j ].data_original );
	}
    }
    //now apply selected cheats
    for( i = 0; i < num_ggp; i++ )
    {
	if( !ggpEntry[ i ].isEnabled )//skip comments and disabled codes
	    continue;

	u8 j;
	for( j = 0; j < ggpEntry[ i ].numCodes; j++ )
	{
	    patch16( ggpEntry[ i ].code[ j ].address, ggpEntry[ i ].code[ j ].data );
	}
    }
}

void GetGGEntries ( char* path )
{
    //gprintf("GetGGEntries( \"%s\" )\n", path );

    if( !num_ggp )
    {
	s32 ret = ReadGGPfile( path );
	if( ret < 0 )
	{
	    //gprintf("ReadGGPfile(): %d\n", ret );

	    num_ggp = 1;

	    snprintf( ggpEntry[ 0 ].text, sizeof( ggpEntry[ 0 ].text ), "Couldn\'t open \"%s\"", path );
	}
    }

    CheatSelector();
    ApplyCheats();
}

/*****************************************************************************/
/*  GUI Buttons data                                                         */
/*****************************************************************************/
static butn_data arrow_up_data =
{
  {NULL,NULL},
  {Button_up_png,Button_up_over_png}
};

static butn_data arrow_down_data =
{
  {NULL,NULL},
  {Button_down_png,Button_down_over_png}
};

/*****************************************************************************/
/*  GUI Arrows button                                                        */
/*****************************************************************************/

static gui_butn arrow_up = {&arrow_up_data,BUTTON_VISIBLE|BUTTON_ACTIVE|BUTTON_OVER_SFX,{0,0,0,0},14,76,436,32};
static gui_butn arrow_down = {&arrow_down_data,BUTTON_VISIBLE|BUTTON_ACTIVE|BUTTON_OVER_SFX,{0,0,0,0},14,368,436,32};

/*****************************************************************************/
/*  GUI helpers                                                              */
/*****************************************************************************/
static gui_item action_cancel =
{
  NULL,Key_B_png,"","Back",10,422,28,28
};

static gui_item action_select =
{
  NULL,Key_A_png,"","Toggle Cheat",602,422,28,28
};

/*****************************************************************************/
/*  GUI Background images                                                    */
/*****************************************************************************/
static gui_image bg_cheat[7] =
{
  {NULL,Bg_main_png,IMAGE_VISIBLE,374,140,284,288,255},
  {NULL,Bg_overlay_png,IMAGE_VISIBLE|IMAGE_REPEAT,0,0,640,480,255},
  {NULL,Banner_top_png,IMAGE_VISIBLE,0,0,640,108,255},
  {NULL,Banner_bottom_png,IMAGE_VISIBLE,0,380,640,100,255},
  {NULL,Main_logo_png,IMAGE_VISIBLE,466,40,152,44,255},
  {NULL,Frame_s1_png,IMAGE_VISIBLE,8,70,450,336,150},
  {NULL,Frame_s2_png,0,384,264,248,140,200},
};

/*****************************************************************************/
/*  GUI Descriptor                                                           */
/*****************************************************************************/
static gui_menu menu_cheat =
{
  "Cheat Menu",
  -1,-1,
  0,0,7,0,
  NULL,
  NULL,
  bg_cheat,
  {&action_cancel, &action_select},
  {&arrow_up,&arrow_down},
  selector_cb
};


static void selector_cb(void)
{
  int i;
  char text[ 256 ];
  int yoffset = PAGEOFFSET;

  /* Initialize directory icon */
  gui_image star_icon;
  star_icon.texture = gxTextureOpenPNG(Star_empty_png,0);
  star_icon.w = star_icon.texture->width;
  star_icon.h = star_icon.texture->height;
  star_icon.x = 24;
  star_icon.y = PAGEOFFSET;

  /* Initialize selection bar */
  gui_image bar_over;
  bar_over.texture = gxTextureOpenPNG(Overlay_bar_png,0);
  //bar_over.w = bar_over.texture->width;
  bar_over.w = 416;
  bar_over.h = bar_over.texture->height;
  bar_over.x = 22;
  bar_over.y = ( -(bar_over.h - star_icon.h)/2 );

  /* Draw Files list */
  for (i = offset; i < (offset + PAGESIZE) && (i < num_ggp); i++)
  {
    if (i == selection)
    {
      /* selection bar */
      gxDrawTexture(bar_over.texture,bar_over.x,yoffset+bar_over.y,bar_over.w,bar_over.h,255);

      /* scrolling text */
      if ( ( string_offset / 14 ) >= strlen( ggpEntry[ i ].text ) )
	string_offset = 0;

      sprintf(text, "%s   ",ggpEntry[ i ].text + string_offset/14);
      strncat(text, ggpEntry[ i ].text, string_offset/14);

      if ( ggpEntry[ i ].isEnabled )
      {
	// directory icon
	gxDrawTexture(star_icon.texture,star_icon.x-1,yoffset-1,star_icon.w+2,star_icon.h+2,255);
	if (FONT_write(text,18,star_icon.x+star_icon.w+6,yoffset+16,bar_over.w-star_icon.w-14,(GXColor){ 126, 122, 103, 0xff }))
	{
	  // string is too large -> scroll text
	  string_offset ++;
	}
      }
      else
      {
	  if( ggpEntry[ i ].type )
	  {
	      if (FONT_write(text,18,star_icon.x+star_icon.w+6,yoffset+16,bar_over.w-star_icon.w-14,(GXColor)WHITE))
	      {
		// text scrolling
		string_offset ++;
	    }
	  }
	  else
	  {
	      if (FONT_write(text,18,star_icon.x+star_icon.w+6,yoffset+16,bar_over.w-star_icon.w-14,(GXColor){ 126, 122, 103, 0xff }))
	      {
	      // text scrolling
	      string_offset ++;
		}
	  }
      }
    }
    else
    {
      if ( ggpEntry[ i ].isEnabled || !ggpEntry[ i ].type )
      {
	if( ggpEntry[ i ].type )
	    gxDrawTexture(star_icon.texture,star_icon.x,yoffset,star_icon.w,star_icon.h,255);

	FONT_write(ggpEntry[ i ].text,16,star_icon.x+star_icon.w+6,yoffset+16,bar_over.w-star_icon.w-14,(GXColor){ 126, 122, 103, 0xff });
      }
      else
      {
	  //FONT_write(ggpEntry[ i ].text,16,star_icon.x,yoffset+16,bar_over.w-8,(GXColor)WHITE);
	  FONT_write(ggpEntry[ i ].text,16,star_icon.x+star_icon.w+6,yoffset+16,bar_over.w-star_icon.w-14,(GXColor)WHITE);
      }

    }

    yoffset += 22;
  }

  gxTextureClose(&bar_over.texture);
  gxTextureClose(&star_icon.texture);
}


void CheatSelector()
{
    //gprintf("CheatSelector()\n\toff: %d\n\tsel: %d\n", offset, selection );
  short p;
  int old = -1;

#ifdef HW_RVL
  int x,y,i,yoffset;
  gui_butn *button;
#endif

  /* background type */
  if (config.bg_type > 0)
  {
    bg_cheat[0].state &= ~IMAGE_REPEAT;
    if (config.bg_type > 1) bg_cheat[0].data = Bg_main_png;
    else bg_cheat[0].data = Bg_main_2_png;
    bg_cheat[0].x = 374;
    bg_cheat[0].y = 140;
    bg_cheat[0].w = 284;
    bg_cheat[0].h = 288;
  }
  else
  {
    bg_cheat[0].state |= IMAGE_REPEAT;
    bg_cheat[0].data = Bg_layer_png;
    bg_cheat[0].x = 0;
    bg_cheat[0].y = 0;
    bg_cheat[0].w = 640;
    bg_cheat[0].h = 480;
  }

  /* background overlay */
  if (config.bg_overlay) bg_cheat[1].state |= IMAGE_VISIBLE;
  else bg_cheat[1].state &= ~IMAGE_VISIBLE;

  /* Initialize Menu */
  gui_menu *m = &menu_cheat;
  GUI_InitMenu(m);
  string_offset = 0;

  while (1)
  {
      if( selection != old )
      {
	  old = selection;
	  string_offset = 0;
      }
	/* update helper */
	if ( m->selected != -1 || !ggpEntry[ selection ].type )
	{
	  /* out of focus */
	  strcpy(action_select.comment,"");
	}
	else
	{
	    if (ggpEntry[ selection ].isEnabled )
		strcpy(action_select.comment,"Disable code");
	    else
		strcpy(action_select.comment,"Enable Code");
	}

    /* Draw menu*/
    GUI_DrawMenu(m);

#ifdef HW_RVL
    if (Shutdown)
    {
      gxTextureClose(&w_pointer);
      GUI_DeleteMenu(m);
      GUI_FadeOut();
      shutdown();
      SYS_ResetSystem(SYS_POWEROFF, 0, 0);
    }
    else if (m_input.ir.valid)
    {
      /* get cursor position */
      x = m_input.ir.x;
      y = m_input.ir.y;

      /* draw wiimote pointer */
      gxDrawTextureRotate(w_pointer, x-w_pointer->width/2, y-w_pointer->height/2, w_pointer->width, w_pointer->height,m_input.ir.angle,255);

      /* find selected item */
      yoffset = PAGEOFFSET  - 4;
      m->selected = m->max_buttons + 2;
      for (i = offset; i < (offset + PAGESIZE) && (i < num_ggp); i++)
      {
	if ((x<=436)&&(y>=yoffset)&&(y<(yoffset+24)))
	{
	  selection = i;
	  m->selected = -1;
	  break;
	}
	yoffset += 24;
      }

      /* find selected button */
      for (i=0; i<2; i++)
      {
	button = m->arrows[i];
	if (button)
	{
	  if (button->state & BUTTON_VISIBLE)
	  {
	    if ((x>=button->x)&&(x<=(button->x+button->w))&&(y>=button->y)&&(y<=(button->y+button->h)))
	    {
	      m->selected = m->max_buttons + i;
	      break;
	    }
	  }
	}
      }
    }
    else
    {
      /* reset indicator */
      m->selected = -1;
    }
#endif

    /* copy EFB to XFB */
    gxSetScreen();

    p = m_input.keys;

    /* highlight next item */
    if (p & PAD_BUTTON_DOWN)
    {
      selection++;
      if (selection >= num_ggp)
	selection = offset = 0;
      if ((selection - offset) >= PAGESIZE)
	offset += PAGESIZE;
    }

    /* highlight previous item */
    else if (p & PAD_BUTTON_UP)
    {
      selection--;
      if (selection < 0)
      {
	selection = num_ggp - 1;
	offset = selection - PAGESIZE + 1;
      }
      if (selection < offset)
	offset -= PAGESIZE;
      if (offset < 0)
	offset = 0;
    }

    /* go back one page */
    else if (p & PAD_TRIGGER_L)
    {
      selection -= PAGESIZE;
      if (selection < 0)
      {
	selection = num_ggp - 1;
	offset = selection - PAGESIZE + 1;
      }
      if (selection < offset)
	offset -= PAGESIZE;
      if (offset < 0)
	offset = 0;
    }

    /* go forward one page */
    else if (p & PAD_TRIGGER_R)
    {
      selection += PAGESIZE;
      if (selection > num_ggp - 1)
	selection = offset = 0;
      if ((selection - offset) >= PAGESIZE)
	offset += PAGESIZE;
    }

    /* quit */
    else if ( ( p & PAD_TRIGGER_Z ) || ( p & PAD_BUTTON_B ) )
    {
	//gprintf("exit cheater screen\n");
	GUI_DeleteMenu(m);
	return;
    }

    /* open selected file or directory */
    else if ( p & PAD_BUTTON_A )
    {
      string_offset = 0;

#ifdef HW_RVL
	/* arrow buttons selected */
	if (m->selected == m->max_buttons) /* up arrow */
	{
	  selection--;
	  if (selection < 0)
	  {
	    selection = num_ggp - 1;
	    offset = selection - PAGESIZE + 1;
	  }
	  if (selection < offset)
	    offset -= PAGESIZE;
	  if (offset < 0)
	    offset = 0;
	}
	else if (m->selected == (m->max_buttons+1)) /* down arrow */
	{
	  selection++;
	  if (selection == num_ggp)
	    selection = offset = 0;
	  if ((selection - offset) >= PAGESIZE)
	    offset += PAGESIZE;
	}

#endif

      /* ensure we are in focus area */
      if ( m->selected < m->max_buttons )
      {
	  if( !ggpEntry[ selection ].type )
	  {
	      //gprintf("clicked a comment\n");
	  }
	  else if( ggpEntry[ selection ].isEnabled )
	  {
	      ggpEntry[ selection ].isEnabled = 0;
	      //gprintf("turned off \"%s\"\n", ggpEntry[ selection ].text );
	  }
	  else if( !ggpEntry[ selection ].isEnabled )
	  {
	      ggpEntry[ selection ].isEnabled = 1;
	      //gprintf("turned on \"%s\"\n", ggpEntry[ selection ].text );
	  }

      }
    }
  }
}

